description = "Write a sliced file in analysis, one file in synthesis, and return some data."


def analysis(sliceno, job):
	# save one file per analysis process...
	filename = 'myfile1'
	data = 'This is job %s analysis slice %d.' % (job, sliceno,)
	# We need "temp=False" for the files written in analysis
	# to be kept permanently.
	# If not set, the files will be erased at job completion.
	job.save(data, filename, sliceno=sliceno, temp=False)


def synthesis(job):
	# ...and one file in the synthesis process...
	filename = 'myfile2'
	data = 'this is job %s synthesis' % (job,)
	job.save(data, filename)

	# ...and let's return some data too.
	returndata = 'this is job %s return value' % (job,)
	return returndata



from .printer import prt

description = "Jobs: Build a job, load its returned value."


def main(urd):
	prt.source(__file__)
	prt()
	prt('Run method "example_returnhelloworld"')
	urd.build('example_returnhelloworld')

	prt()
	prt('Run method "example_returnhelloworld" again')
	job = urd.build('example_returnhelloworld')
	prt('Note that is is linked, and not re-executed.')

	prt()
	prt('Load and print the job\'s return value')
	result = job.load()
	prt.output(result)

	prt()
	prt('Try these commands:')
	prt.command('ax job', job)
	prt.command('ls', job.path)



############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2018-2022 Carl Drougge                       #
# Modifications copyright (c) 2020-2021 Anders Berkeman                    #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division

from threading import Thread
import multiprocessing
import signal
from os import unlink
from os.path import join
import time

from accelerator import dependency
from accelerator import dispatch

from accelerator import workspace
from accelerator import database
from accelerator import methods
from accelerator.colourwrapper import colour
from accelerator.setupfile import update_setup
from accelerator.job import WORKDIRS, Job
from accelerator.extras import json_save, DotDict
from accelerator.error import BuildError




class Main:
	""" This is the main controller behind the server. """

	def __init__(self, config, options, server_url):
		"""
		Setup objects:

		  Methods

		  WorkSpaces

		"""
		self.config = config
		self.keep_temp_files = options.keep_temp_files
		self.server_url = server_url
		self._update_methods()
		self.target_workdir = self.config['target_workdir']
		self.workspaces = {}
		for name, path in self.config.workdirs.items():
			self.workspaces[name] = workspace.WorkSpace(name, path, config.slices)
		WORKDIRS.clear()
		WORKDIRS.update({k: v.path for k, v in self.workspaces.items()})
		self.DataBase = database.DataBase(self)
		self.update_database()
		self.broken = False

	def _update_methods(self):
		print('Update methods')
		# initialise methods class looking in method_directories from config file
		self.Methods = methods.Methods(self.config)

	def update_methods(self):
		try:
			self._update_methods()
			self.update_database()
			self.broken = False
		except methods.MethodLoadException as e:
			self.broken = e.module_list
			return {'broken': e.module_list}


	def get_workspace_details(self):
		""" Some information about main workspace, some parts of config """
		return dict(
			[(key, getattr(self.workspaces[self.target_workdir], key),) for key in ('slices',)] +
			[(key, self.config.get(key),) for key in ('input_directory', 'result_directory', 'common_directory', 'urd',)]
		)


	def list_workdirs(self):
		""" Return list of all initiated workdirs """
		return self.workspaces


	def print_workdirs(self):
		namelen = max(len(n) for n in self.workspaces)
		templ = "    %%s %%%ds: %%s %s(%%d)" % (namelen, colour.RESET,)
		print("Available workdirs:")
		names = sorted(self.workspaces)
		names.remove(self.target_workdir)
		names.insert(0, self.target_workdir)
		for n in names:
			if n == self.target_workdir:
				prefix = 'TARGET  ' + colour.BOLD
			else:
				prefix = 'SOURCE  '
			w = self.workspaces[n]
			print(templ % (prefix, n, w.path, w.slices,))


	def add_single_jobid(self, jobid):
		ws = self.workspaces[jobid.rsplit('-', 1)[0]]
		ws.add_single_jobid(jobid)
		return self.DataBase.add_single_jobid(jobid)

	def update_database(self):
		"""Insert all new jobids (from all workdirs) in database,
		discard all deleted or with incorrect hash.
		"""
		if hasattr(multiprocessing, 'get_context') and not running_under_wsl():
			# forkserver is apparently broken under (some versions of) WSL1
			ctx = multiprocessing.get_context('forkserver')
			Pool = ctx.Pool
		else:
			Pool = multiprocessing.Pool
		pool = Pool(initializer=_pool_init, initargs=(WORKDIRS,))
		try:
			self.DataBase._update_begin()
			def update(name):
				ws = self.workspaces[name]
				ws.update(pool)
				self.DataBase._update_workspace(ws, pool)
				failed.remove(name)
			failed = set(self.workspaces)
			t_l = []
			for name in self.workspaces:
				# Run all updates in parallel. This gets all (sync) listdir calls
				# running at the same time. Then each workspace will use the same
				# process pool to do the post.json checking.
				t = Thread(
					target=update,
					args=(name,),
					name='Update ' + name,
				)
				t.daemon = True
				t.start()
				t_l.append(t)
			for t in t_l:
				t.join()
			assert not failed, "%s failed to update" % (', '.join(failed),)
		finally:
			pool.close()
		self.DataBase._update_finish(self.Methods.hash)


	def initialise_jobs(self, setup, workdir=None):
		""" Updata database, check deps, create jobids. """
		ws = workdir or self.target_workdir
		if ws not in self.workspaces:
			raise BuildError("Workdir %s does not exist" % (ws,))
		return dependency.initialise_jobs(
			setup,
			self.workspaces[ws],
			self.DataBase,
			self.Methods,
		)


	def run_job(self, jobid, subjob_cookie=None, parent_pid=0, concurrency=None):
		W = self.workspaces[Job(jobid).workdir]
		#
		active_workdirs = {name: ws.path for name, ws in self.workspaces.items()}
		slices = self.workspaces[self.target_workdir].slices

		t0 = time.time()
		setup = update_setup(jobid, starttime=t0)
		prof = setup.get('exectime', DotDict())
		new_prof, files, subjobs = dispatch.launch(W.path, setup, self.config, self.Methods, active_workdirs, slices, concurrency, self.server_url, subjob_cookie, parent_pid)
		prefix = join(W.path, jobid) + '/'
		if not self.keep_temp_files:
			for filename, temp in list(files.items()):
				if temp:
					unlink(join(prefix, filename))
					del files[filename]
		prof.update(new_prof)
		prof.total = 0
		prof.total = sum(v for v in prof.values() if isinstance(v, (float, int)))
		if concurrency:
			prof.concurrency = concurrency
		data = dict(
			starttime=t0,
			endtime=time.time(),
			exectime=prof,
		)
		update_setup(jobid, **data)
		data['files'] = sorted(fn[len(prefix):] if fn.startswith(prefix) else fn for fn in files)
		data['subjobs'] = subjobs
		data['version'] = 1
		json_save(data, jobid.filename('post.json'))


	def get_methods(self):
		return {k: self.method_info(k) for k in self.Methods.db}


	def method_info(self, method):
		d = self.Methods.db.get(method, None)
		if d:
			d = dict(d)
			p = self.Methods.params[method]
			for k in ('options', 'datasets', 'jobs'):
				d[k] = [v[0] if isinstance(v, (list, tuple)) else v for v in p[k]]
			d['description'] = self.Methods.descriptions[method]
			return d


def _pool_init(workdirs):
	# The pool system will send SIGTERM when the pool is closed, so
	# restore the original behaviour for that.
	signal.signal(signal.SIGTERM, signal.SIG_DFL)
	WORKDIRS.update(workdirs)


# This can't distinguish between WSL1 and WSL2.
_running_under_wsl = None
def running_under_wsl():
	global _running_under_wsl
	if _running_under_wsl is None:
		try:
			with open('/proc/sys/kernel/osrelease', 'rb') as fh:
				osrelease = fh.read().lower()
			_running_under_wsl = (b'microsoft' in osrelease or b'wsl' in osrelease)
		except IOError:
			_running_under_wsl = False
	return _running_under_wsl



############################################################################
#                                                                          #
# Copyright (c) 2022 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Verify that datasets can not be overwritten.
'''

from accelerator.error import DatasetUsageError

def synthesis(job):
	dw = job.datasetwriter(name='a', columns={'a': 'ascii'})
	try:
		job.datasetwriter(name='a')
		raise Exception("Allowed creating a duplicate datasetwriter")
	except DatasetUsageError:
		pass
	dw.get_split_write()
	a = dw.finish()
	try:
		job.datasetwriter(name='a')
		raise Exception("Allowed creating a datasetwriter for an existing dataset")
	except DatasetUsageError:
		pass
	dw = job.datasetwriter(name='b', columns={'b': 'ascii'})
	try:
		a.link_to_here('b')
		raise Exception("Allowed link_to_here over a datasetwriter")
	except DatasetUsageError:
		pass
	dw.get_split_write()
	dw.finish()
	try:
		a.link_to_here('b')
		raise Exception("Allowed link_to_here over an existing dataset")
	except DatasetUsageError:
		pass



from .printer import prt

description = "Dataset: Append columns to an existing dataset."


def main(urd):
	prt.source(__file__)
	prt()
	prt('Create a dataset.')
	job = urd.build('dsexample_createdataset')

	prt()
	prt('Append a new columnd to the dataset.')
	job = urd.build('dsexample_appendcolumn', source=job)

	prt()
	prt('Try')
	prt.command('ax ds', job)
	prt('The "parent" row shows that this dataset includes columns')
	prt('from a parent dataset.')
	prt()
	prt('To see the location of the different columns, try')
	prt.command('ax ds -w', job)



############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2018-2022 Carl Drougge                       #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division

from accelerator import _dsutil
from accelerator.compat import str_types, PY3

_convfuncs = {
	'number'   : _dsutil.WriteNumber,
	'complex64': _dsutil.WriteComplex64,
	'complex32': _dsutil.WriteComplex32,
	'float64'  : _dsutil.WriteFloat64,
	'float32'  : _dsutil.WriteFloat32,
	'int64'    : _dsutil.WriteInt64,
	'int32'    : _dsutil.WriteInt32,
	'bits64'   : _dsutil.WriteBits64,
	'bits32'   : _dsutil.WriteBits32,
	'bool'     : _dsutil.WriteBool,
	'datetime' : _dsutil.WriteDateTime,
	'date'     : _dsutil.WriteDate,
	'time'     : _dsutil.WriteTime,
	'bytes'    : _dsutil.WriteBytes,
	'ascii'    : _dsutil.WriteAscii,
	'unicode'  : _dsutil.WriteUnicode,
	'parsed:number'   : _dsutil.WriteParsedNumber,
	'parsed:complex64': _dsutil.WriteParsedComplex64,
	'parsed:complex32': _dsutil.WriteParsedComplex32,
	'parsed:float64'  : _dsutil.WriteParsedFloat64,
	'parsed:float32'  : _dsutil.WriteParsedFloat32,
	'parsed:int64'    : _dsutil.WriteParsedInt64,
	'parsed:int32'    : _dsutil.WriteParsedInt32,
	'parsed:bits64'   : _dsutil.WriteParsedBits64,
	'parsed:bits32'   : _dsutil.WriteParsedBits32,
}

_type2iter = {
	'number'   : _dsutil.ReadNumber,
	'complex64': _dsutil.ReadComplex64,
	'complex32': _dsutil.ReadComplex32,
	'float64'  : _dsutil.ReadFloat64,
	'float32'  : _dsutil.ReadFloat32,
	'int64'    : _dsutil.ReadInt64,
	'int32'    : _dsutil.ReadInt32,
	'bits64'   : _dsutil.ReadBits64,
	'bits32'   : _dsutil.ReadBits32,
	'bool'     : _dsutil.ReadBool,
	'datetime' : _dsutil.ReadDateTime,
	'date'     : _dsutil.ReadDate,
	'time'     : _dsutil.ReadTime,
	'bytes'    : _dsutil.ReadBytes,
	'ascii'    : _dsutil.ReadAscii,
	'unicode'  : _dsutil.ReadUnicode,
}

def typed_writer(typename):
	if typename not in _convfuncs:
		raise ValueError("Unknown writer for type %s" % (typename,))
	return _convfuncs[typename]

def typed_reader(typename):
	if typename not in _type2iter:
		raise ValueError("Unknown reader for type %s" % (typename,))
	return _type2iter[typename]

_nodefault = object()

from json import JSONEncoder, JSONDecoder, loads as json_loads
class WriteJson(object):
	__slots__ = ('fh', 'encode')
	min = max = None
	def __init__(self, *a, **kw):
		default = kw.pop('default', _nodefault)
		if PY3:
			self.fh = _dsutil.WriteUnicode(*a, **kw)
			self.encode = JSONEncoder(ensure_ascii=False, separators=(',', ':')).encode
		else:
			self.fh = _dsutil.WriteBytes(*a, **kw)
			self.encode = JSONEncoder(ensure_ascii=True, separators=(',', ':')).encode
		self.encode = self._wrap_encode(self.encode, default)
	def _wrap_encode(self, encode, default):
		if default is _nodefault:
			return encode
		default = encode(default)
		def wrapped_encode(o):
			try:
				return encode(o)
			except (TypeError, ValueError):
				return default
		return wrapped_encode
	def write(self, o):
		self.fh.write(self.encode(o))
	@property
	def count(self):
		return self.fh.count
	@property
	def compression(self):
		return self.fh.compression
	def close(self):
		self.fh.close()
	def __enter__(self):
		return self
	def __exit__(self, type, value, traceback):
		self.close()
_convfuncs['json'] = WriteJson

class WriteParsedJson(WriteJson):
	"""This assumes strings are the object you wanted and parse them as json.
	If they are unparseable you get an error."""
	__slots__ = ()
	def _wrap_encode(self, encode, default):
		if default is not _nodefault:
			if isinstance(default, str_types):
				default = json_loads(default)
			default = encode(default)
		def wrapped_encode(o):
			try:
				if isinstance(o, str_types):
					o = json_loads(o)
				return encode(o)
			except (TypeError, ValueError):
				if default is _nodefault:
					raise
				return default
		return wrapped_encode
_convfuncs['parsed:json'] = WriteParsedJson

class ReadJson(object):
	__slots__ = ('fh', 'decode')
	def __init__(self, *a, **kw):
		if PY3:
			self.fh = _dsutil.ReadUnicode(*a, **kw)
		else:
			self.fh = _dsutil.ReadBytes(*a, **kw)
		self.decode = JSONDecoder().decode
	def __next__(self):
		return self.decode(next(self.fh))
	next = __next__
	def close(self):
		self.fh.close()
	def __iter__(self):
		return self
	def __enter__(self):
		return self
	def __exit__(self, type, value, traceback):
		self.close()
_type2iter['json'] = ReadJson

from pickle import dumps as pickle_dumps, loads as pickle_loads
class WritePickle(object):
	__slots__ = ('fh',)
	min = max = None
	def __init__(self, *a, **kw):
		assert PY3, "Pickle columns require python 3, sorry"
		assert 'default' not in kw, "default not supported for Pickle, sorry"
		self.fh = _dsutil.WriteBytes(*a, **kw)
	def write(self, o):
		self.fh.write(pickle_dumps(o, 4))
	@property
	def count(self):
		return self.fh.count
	@property
	def compression(self):
		return self.fh.compression
	def close(self):
		self.fh.close()
	def __enter__(self):
		return self
	def __exit__(self, type, value, traceback):
		self.close()
_convfuncs['pickle'] = WritePickle

class ReadPickle(object):
	__slots__ = ('fh',)
	def __init__(self, *a, **kw):
		assert PY3, "Pickle columns require python 3, sorry"
		self.fh = _dsutil.ReadBytes(*a, **kw)
	def __next__(self):
		return pickle_loads(next(self.fh))
	next = __next__
	def close(self):
		self.fh.close()
	def __iter__(self):
		return self
	def __enter__(self):
		return self
	def __exit__(self, type, value, traceback):
		self.close()
_type2iter['pickle'] = ReadPickle



from .printer import prt

description = "Dataset: Create a dataset from scratch."


def main(urd):
	prt.source(__file__)
	prt()
	prt('Build a job that creates a dataset from scratch.')
	job = urd.build('dsexample_createdataset')

	prt()
	prt('Check contents of dataset using')
	prt.command('ax cat -H', job)

	prt()
	prt('More info about the dataset')
	prt.command('ax ds -s', job)

	prt()
	prt('For convenience, the jobid can be used as a reference to the default')
	prt('dataset in a job.  The full name is "%s/default".' % (job,))



from accelerator import Job
from .printer import prt

description = "Jobs: Subjob"


def main(urd):
	prt.source(__file__)
	prt()
	prt('This job will execute a subjob')
	job = urd.build('example_buildsubjob')

	prt()
	prt('Print subjob info')
	for sjob in job.post.subjobs:
		prt.plain('subjob', dict(job=sjob, method=Job(sjob).method))
	prt()
	prt('You can see subjobs in the output from')
	prt.command('ax job', job)



############################################################################
#                                                                          #
# Copyright (c) 2021 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test dataset_fanout with varying types, hashlabel and chain truncation.
'''

from accelerator import subjobs
from accelerator.compat import unicode

from itertools import cycle

def synthesis(job):
	def mk(name, types, lines, hashlabel=None, previous=None):
		columns = {chr(ix): typ for ix, typ in enumerate(types, 65)}
		dw = job.datasetwriter(name=name, columns=columns, hashlabel=hashlabel, previous=previous)
		w = dw.get_split_write_list()
		for line in lines:
			w(line)
		return dw.finish()

	def chk(job, colnames, types, ds2lines, previous={}, hashlabel=None):
		have_ds = set(ds.name for ds in job.datasets)
		want_ds = set(ds2lines)
		assert have_ds == want_ds, 'Job %r should have had datasets %r but had %r' % (job, want_ds, have_ds,)
		colnames = sorted(colnames)
		for ds, lines in ds2lines.items():
			ds = job.dataset(ds)
			assert ds.hashlabel == hashlabel, 'Dataset %s should have had hashlabel %s but had %s' % (ds.quoted, hashlabel, ds.hashlabel,)
			assert ds.previous == previous.get(ds.name), 'Dataset %s should have had previous %s but had %s' % (ds.quoted, previous.get(ds.name), ds.previous,)
			ds_colnames = sorted(ds.columns)
			assert ds_colnames == colnames, 'Dataset %s should have had columns %r but had %r' % (ds.quoted, colnames, ds_colnames,)
			ds_types = tuple(col.type for _, col in sorted(ds.columns.items()))
			assert ds_types == types, 'Dataset %s should have had columns with types %r but had %r' % (ds.quoted, types, ds_types,)
			have_lines = sorted(ds.iterate(None))
			want_lines = sorted(lines)
			assert have_lines == want_lines, 'Dataset %s should have contained %r but contained %r' % (ds.quoted, want_lines, have_lines,)

	# just a simple splitting
	a = mk('a', ('unicode', 'ascii', 'int64'), [('a', 'a', 1), ('b', 'b', 2), ('a', 'c', 3)], hashlabel='A')
	j_a_A = subjobs.build('dataset_fanout', source=a, column='A')
	chk(j_a_A, 'BC', ('ascii', 'int64'), {'a': [('a', 1), ('c', 3)], 'b': [('b', 2)]})
	j_a_B = subjobs.build('dataset_fanout', source=a, column='B')
	chk(j_a_B, 'AC', ('unicode', 'int64'), {'a': [('a', 1)], 'b': [('b', 2)], 'c': [('a', 3)]}, hashlabel='A')

	# non-text columns should work too
	j_a_C = subjobs.build('dataset_fanout', source=a, column='C')
	chk(j_a_C, 'AB', ('unicode', 'ascii'), {'1': [('a', 'a')], '2': [('b', 'b')], '3': [('a', 'c')]}, hashlabel='A')

	b = mk('b', ('ascii', 'unicode', 'int32', 'int32'), [('a', 'aa', 11, 111), ('b', 'bb', 12, 112), ('a', 'cc', 13, 113), ('d', 'dd', 14, 114)], previous=a)
	# with previous
	j_b_A = subjobs.build('dataset_fanout', source=b, column='A', previous=j_a_A)
	chk(
		j_b_A,
		'BCD',
		('unicode', 'int32', 'int32'),
		{'a': [('aa', 11, 111), ('cc', 13, 113)], 'b': [('bb', 12, 112)], 'd': [('dd', 14, 114)]},
		previous={'a': j_a_A.dataset('a'), 'b': j_a_A.dataset('b')},
	)

	# without previous, but only getting the data from b because of length=1
	j_b_A_len1 = subjobs.build('dataset_fanout', source=b, column='A', length=1)
	chk(
		j_b_A_len1,
		'BCD',
		('unicode', 'int32', 'int32'),
		{'a': [('aa', 11, 111), ('cc', 13, 113)], 'b': [('bb', 12, 112)], 'd': [('dd', 14, 114)]},
	)

	# with "wrong" previous, inheriting some empty datasets.
	j_b_A_C = subjobs.build('dataset_fanout', source=b, column='A', previous=j_a_C)
	chk(
		j_b_A_C,
		'BCD',
		('unicode', 'int32', 'int32'),
		{'a': [('aa', 11, 111), ('cc', 13, 113)], 'b': [('bb', 12, 112)], 'd': [('dd', 14, 114)], '1': [], '2': [], '3': []},
		previous={'1': j_a_C.dataset('1'), '2': j_a_C.dataset('2'), '3': j_a_C.dataset('3')},
	)

	# without previous, getting data from both a and b and the "widest" type for the columns.
	# (discards the D column since it doesn't exist in a.)
	j_b_A_None = subjobs.build('dataset_fanout', source=b, column='A')
	chk(
		j_b_A_None,
		'BC',
		('unicode', 'int64'),
		{'a': [('a', 1), ('aa', 11), ('c', 3), ('cc', 13)], 'b': [('b', 2), ('bb', 12)], 'd': [('dd', 14)]},
	)

	# test more type combinations, and switching hashlabel (to an included column)
	tt_a = mk(
		'tt_a',
		('ascii', 'int32', 'bits64', 'float32', 'number', 'complex32', 'number'),
		[('a', 1, 2, 2.5, 3, 1+2j, 3.14)],
		hashlabel='B',
	)
	tt_b = mk(
		'tt_b',
		('ascii', 'int64', 'bits32', 'float64', 'int32', 'complex64', 'float64'),
		[('a', 11, 12, 12.5, 13, 11+2j, 13.14)],
		hashlabel='B',
		previous=tt_a,
	)
	tt_c = mk(
		'tt_c',
		('ascii', 'int32', 'bits64', 'int64', 'float64', 'complex32', 'float32'),
		[('a', 111, 112, 112, 113.5, 111+2j, 314.0), ('b', 0, 0, 0, 0, 0, 0)],
		hashlabel='C',
		previous=tt_b,
	)

	# first two, some type changes
	j_tt_b = subjobs.build('dataset_fanout', source=tt_b, column='A')
	chk(
		j_tt_b,
		'BCDEFG',
		('int64', 'bits64', 'float64', 'number', 'complex64', 'number'),
		{'a': [(1, 2, 2.5, 3, 1+2j, 3.14), (11, 12, 12.5, 13, 11+2j, 13.14)]},
		hashlabel='B',
	)

	# all three in one, more types become number
	j_tt_c = subjobs.build('dataset_fanout', source=tt_c, column='A')
	chk(
		j_tt_c,
		'BCDEFG',
		('int64', 'bits64', 'number', 'number', 'complex64', 'number'),
		{'a': [(1, 2, 2.5, 3, 1+2j, 3.14), (11, 12, 12.5, 13, 11+2j, 13.14), (111, 112, 112, 113.5, 111+2j, 314.0)], 'b': [(0, 0, 0, 0, 0, 0)]},
		hashlabel=None,
	)

	# just two (checking that earlier types are not considered)
	j_tt_c_len2 = subjobs.build('dataset_fanout', source=tt_c, column='A', length=2)
	chk(
		j_tt_c_len2,
		'BCDEFG',
		('int64', 'bits64', 'number', 'number', 'complex64', 'float64'),
		{'a': [(11, 12, 12.5, 13, 11+2j, 13.14), (111, 112, 112, 113.5, 111+2j, 314.0)], 'b': [(0, 0, 0, 0, 0, 0)]},
		hashlabel=None,
	)

	# using previous to only get one source dataset, again checking that earlier
	# types are not considered and that only a gets a previous (and b doesn't)
	j_tt_c_b = subjobs.build('dataset_fanout', source=tt_c, column='A', previous=j_tt_b)
	chk(
		j_tt_c_b,
		'BCDEFG',
		('int32', 'bits64', 'int64', 'float64', 'complex32', 'float32'),
		{'a': [(111, 112, 112, 113.5, 111+2j, 314.0)], 'b': [(0, 0, 0, 0, 0, 0)]},
		hashlabel='C',
		previous={'a': j_tt_b.dataset('a')},
	)

	# it generally works, let's make an exhaustive test of compatible types
	# (to check that the values actually are compatible)
	previous = None
	all_types = []
	want_data = []
	for ix, types in enumerate(zip(
		cycle(['ascii']), # this is the split column
		['bits32', 'bits64', 'int32', 'int64', 'float32', 'float64', 'number'],
		cycle(['bits64', 'bits32']),
		cycle(['complex64', 'complex32']),
		cycle(['float64', 'float32']),
		cycle(['int64', 'bits32', 'int32']),
		cycle(['unicode', 'ascii']),
	)):
		data = [('data',) + (ix + 1000,) * 5 + (unicode(ix),)]
		want_data.append(data[0][1:])
		all_types.append(
			mk('all types %d' % (ix,), types, data, previous=previous)
		)
		previous = all_types[-1]

	j_all = subjobs.build('dataset_fanout', source=all_types[-1], column='A')
	chk(
		j_all,
		'BCDEFG',
		('number', 'bits64', 'complex64', 'float64', 'int64', 'unicode'),
		{'data': want_data},
	)

	# the B column doesn't have number any more here, but should still become number.
	j_all_except_number = subjobs.build('dataset_fanout', source=all_types[-2], column='A')
	chk(
		j_all_except_number,
		'BCDEFG',
		('number', 'bits64', 'complex64', 'float64', 'int64', 'unicode'),
		{'data': want_data[:-1]},
	)



from os.path import dirname, join
from .printer import prt

description = "Dataset: Import, type, sort, and hash partition a tabular file."


# file is stored in same directory as this python file
filename = join(dirname(__file__), 'data.csv')


def main(urd):
	prt.source(__file__)
	prt()
	prt('Import, type, sort, and hash partition the file "data.csv":')
	prt()

	imp = urd.build('csvimport', filename=filename)
	imp = urd.build(
		'dataset_type',
		source=imp,
		column2type=dict(
			Date='date:%Y-%m-%d',
			String='unicode:utf-8',
			Int='number',
			Float='float64',
		),
	)
	imp = urd.build('dataset_sort', source=imp, sort_columns='Date')
	imp = urd.build('dataset_hashpart', source=imp, hashlabel='String')

	job = urd.build('dsexample_aggandmergedata', source=imp)

	prt()
	prt('Now you can do')
	prt.command('ax ds', imp)
	prt('to show info on the default dataset in the dataset_hashpart job.  Note that')
	prt('the latest built dataset_hashpart dataset can also be specified like this')
	prt.command('ax ds dataset_hashpart')

	prt()
	prt('To see the output (to stdout) from the "dsexample_aggandmergedata" job, do')
	prt.command('ax job -O', job)
	prt('And, to see information about the job itself, do')
	prt.command('ax job', job)

	prt()
	prt('To see the contents of the dataset, type')
	prt.command('ax cat -H', imp)
	prt('or, for example')
	prt.command('ax cat -f json', imp)

	prt()
	prt('Please check options using "ax ds --help" etc.')



############################################################################
#                                                                          #
# Copyright (c) 2019-2021 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Verify that each slice contains the expected data after test_datasetwriter.
'''

from datetime import date
from sys import version_info

from accelerator.dataset import Dataset
from accelerator.dsutil import typed_writer
from . import test_data

depend_extra=(test_data,)

datasets = ('source',)

def analysis(sliceno, params):
	assert list(datasets.source.iterate(sliceno, "a")) == [sliceno, 42]
	assert list(datasets.source.iterate(sliceno, "b")) == ["a", str(sliceno)]
	named = Dataset(datasets.source, "named")
	assert list(named.iterate(sliceno, "c")) == [True, False]
	assert list(named.iterate(sliceno, "d")) == [date(1536, 12, min(sliceno + 1, 31)), date(2236, 5, min(sliceno + 1, 31))]
	if sliceno < test_data.value_cnt:
		passed = Dataset(datasets.source, "passed")
		good = tuple(v[sliceno] for _, v in sorted(test_data.data.items()))
		assert list(passed.iterate(sliceno)) == [good]
		if version_info > (3, 6, 0):
			want_fold = (sliceno == 1)
			assert next(passed.iterate(sliceno, "datetime")).fold == want_fold
			assert next(passed.iterate(sliceno, "time")).fold == want_fold
	synthesis_split = Dataset(datasets.source, "synthesis_split")
	values = zip((1, 2, 3,), "abc")
	hash = typed_writer("int32").hash
	good = [v for v in values if hash(v[0]) % params.slices == sliceno]
	assert list(synthesis_split.iterate(sliceno)) == good
	synthesis_manual = Dataset(datasets.source, "synthesis_manual")
	assert list(synthesis_manual.iterate(sliceno, "sliceno")) == [sliceno]
	nonetest = Dataset(datasets.source, "nonetest")
	good = tuple(v[0] if k in test_data.not_none_capable else None for k, v in sorted(test_data.data.items()))
	assert list(nonetest.iterate(sliceno)) == [good]



from .printer import prt

description = "Dataset: Several datasets in a job - print some info per dataset"


def main(urd):
	prt.source(__file__)
	prt()
	prt('This job creates three datasets')
	job = urd.build('dsexample_multipledatasets')

	prt()
	prt('Print dataset names, columns and types')
	for ds in job.datasets:
		prt.plain(ds, sum(ds.lines), {name: ds.columns[name].type for name in ds.columns})

	prt()
	prt('To see all datasets in a job, try')
	prt.command('ax ds -l', job)

	prt()
	prt('Print contents of a specific dataset like this')
	prt.command('ax cat -H %s/third' % (job,))



############################################################################
#                                                                          #
# Copyright (c) 2021 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test the "ax ds" shell command. Primarily tests the job spec parser.
'''

options = dict(
	command_prefix=['ax', '--config', '/some/path/here'],
	want={'spec': 'ds'},
)

import os
from subprocess import check_output

def ax_ds(*a):
	cmd = options.command_prefix + ['ds'] + list(a)
	print(cmd)
	res = check_output(cmd)
	res = res.decode('utf-8', 'replace')
	print(res)
	return res.split('\n')

def synthesis(job):
	os.putenv('XDG_CONFIG_HOME', job.path) # make sure we can't be messed up by config
	for spec, ds in options.want.items():
		res = ax_ds(spec)
		got_ds = res[0]
		assert ds == ds, 'Spec %r should have given %r but gave %r' % (spec, ds, got_ds,)




############################################################################
#                                                                          #
# Copyright (c) 2020-2022 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

from traceback import print_exc
from datetime import datetime
import errno
from argparse import RawTextHelpFormatter
import sys

from accelerator.compat import ArgumentParser
from accelerator.colourwrapper import colour
from accelerator.setupfile import encode_setup
from accelerator.compat import FileNotFoundError, url_quote
from accelerator.unixhttp import call
from .parser import name2job, JobNotFound

def show(url, job, show_output):
	print(job.path)
	print('=' * len(job.path))
	setup = job.json_load('setup.json')
	setup.pop('_typing', None)
	setup.starttime = str(datetime.fromtimestamp(setup.starttime))
	if 'endtime' in setup:
		setup.endtime = str(datetime.fromtimestamp(setup.endtime))
	print(encode_setup(setup, as_str=True))
	if job.datasets:
		print()
		print('datasets:')
		for ds in job.datasets:
			print('   ', ds.quoted)
	try:
		post = job.json_load('post.json')
	except FileNotFoundError:
		print(colour('WARNING: Job did not finish', 'job/warning'))
		post = None
	if post and post.subjobs:
		print()
		print('subjobs:')
		for sj in sorted(post.subjobs):
			print('   ', sj)
	if post and post.files:
		print()
		print('files:')
		for fn in sorted(post.files):
			print('   ', job.filename(fn))
	if post and not call(url + '/job_is_current/' + url_quote(job)):
		print(colour('Job is not current', 'job/info'))
	print()
	out = job.output()
	if show_output:
		if out:
			print('output (use --just-output/-O to see only the output):')
			print(out)
			if not out.endswith('\n'):
				print()
		else:
			print(job, 'produced no output')
			print()
	elif out:
		print('%s produced %d bytes of output, use --output/-o to see it' % (job, len(out),))
		print()

def main(argv, cfg):
	descr = 'show setup.json, dataset list, etc for jobs'
	parser = ArgumentParser(
		prog=argv.pop(0),
		description=descr,
		formatter_class=RawTextHelpFormatter,
	)
	group = parser.add_mutually_exclusive_group()
	group.add_argument('-o', '--output', action='store_true', help='show job output')
	group.add_argument('-O', '--just-output', action='store_true', help='show only job output')
	group.add_argument('-P', '--just-path', action='store_true', help='show only job path')
	parser.add_argument(
		'jobid',
		nargs='+', metavar='jobid/jobspec',
		help='jobid is just a jobid.\n' +
		     'you can also use path, method or :urdlist:[entry].\n' +
		     'path is to a jobdir (with setup.json in it).\n' +
		     'method is the latest (current) job with that method (i.e\n' +
		     'the latest finished job with current source code).\n' +
		     ':urdlist:[entry] looks up jobs in urd. details are in the\n' +
		     'urd help, except here entry defaults to -1 and you can\'t\n' +
		     'list things (no .../ or .../since/x).\n' +
		     'you can use spec~ or spec~N to go back N current jobs\n' +
		     'with that method or spec^ or spec^N to follow .previous'
	)
	args = parser.parse_intermixed_args(argv)
	res = 0
	for path in args.jobid:
		try:
			job = name2job(cfg, path)
			if args.just_output:
				out = job.output()
				if out:
					print(out, end='' if out.endswith('\n') else '\n')
			elif args.just_path:
				print(job.path)
			else:
				show(cfg.url, job, args.output)
		except JobNotFound as e:
			print(e)
			res = 1
		except Exception as e:
			if isinstance(e, OSError) and e.errno == errno.EPIPE:
				raise
			print_exc(file=sys.stderr)
			print("Failed to show %r" % (path,), file=sys.stderr)
			res = 1
	return res



############################################################################
#                                                                          #
# Copyright (c) 2019 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

datasets = ("a", "b",)

def analysis(sliceno):
	assert sorted(datasets.a.columns) == sorted(datasets.b.columns)
	iter_a = datasets.a.iterate(sliceno)
	iter_b = datasets.b.iterate(sliceno, status_reporting=False)
	# This doesn't use zip because then we can't see if one iterator is shorter
	while True:
		try:
			a = next(iter_a)
		except StopIteration:
			try:
				next(iter_b)
				raise Exception("dataset b is longer than a in slice %d" % (sliceno,))
			except StopIteration:
				break
		try:
			b = next(iter_b)
		except StopIteration:
			raise Exception("dataset a is longer than b in slice %d" % (sliceno,))
		assert a == b



datasets = ('source',)

description = "Iterate and aggregate data from a chained dataset."


def analysis(sliceno):
	# this is equivalent to just
	# return list(datasets.source.iterate_chain(sliceno, ['Int', 'String']))
	# but written out so you can see what it does.
	v = []
	for n, s in datasets.source.iterate_chain(sliceno, ['Int', 'String']):
		v.append((n, s))
	return v


def synthesis(analysis_res):
	v = analysis_res.merge_auto()
	for item in v:
		print(item)
	return v



############################################################################
#                                                                          #
# Copyright (c) 2020-2021 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Verify that all column types come out correctly in csvexport.
'''

from datetime import date, time, datetime

from accelerator import subjobs, status
from accelerator.dsutil import _convfuncs
from accelerator.compat import PY2

def synthesis(job):
	dw = job.datasetwriter()
	todo = {
		'ascii',
		'bits32',
		'bits64',
		'bool',
		'bytes',
		'complex32',
		'complex64',
		'date',
		'datetime',
		'float32',
		'float64',
		'int32',
		'int64',
		'json',
		'number',
		'pickle',
		'time',
		'unicode',
	}
	check = {n for n in _convfuncs if not n.startswith('parsed:')}
	assert todo == check, 'Missing/extra column types: %r %r' % (check - todo, todo - check,)
	no_none = {'bits32', 'bits64', 'json'}
	for name in sorted(todo):
		if PY2 and name == 'pickle':
			# pickle columns are not supported on python 2.
			t = 'ascii'
		else:
			t = name
		dw.add(name, t, none_support=name not in no_none)
	write = dw.get_split_write()
	write(
		'a', 0xffffffff, 0xfedcba9876543210, True, b'hello',
		42, 1e100+0.00000000000000001j,
		date(2020, 6, 23), datetime(2020, 6, 23, 12, 13, 14),
		1.0, float('-inf'), -10, -20,
		{'json': True}, 0xfedcba9876543210beef,
		'...' if PY2 else 1+2j, time(12, 13, 14), 'bl\xe5',
	)
	d = {}
	d['recursion'] = d
	write(
		'b', 0, 0, False, b'bye',
		2-3j, -7,
		date(1868,  1,  3), datetime(1868,  1,  3, 13, 14, 5),
		float('inf'), float('nan'), 0, 0,
		[False, None], 42.18,
		'...' if PY2 else d, time(13, 14, 5), 'bl\xe4',
	)
	write(
		None, 72, 64, None, None,
		None, None,
		None, None,
		None, None, None, None,
		None, None,
		None, None, None,
	)
	ds = dw.finish()
	sep = '\x1e'
	for sep, q, none_as, last_line in (
		('\x1e', '', None, (
			'None', '72', '64', 'None', 'None',
			'None', 'None',
			'None', 'None',
			'None', 'None', 'None', 'None',
			'null', 'None',
			'None', 'None', 'None',
		)),
		('\x1e', 'a', '', (
			'', '72', '64', '', '',
			'', '',
			'', '',
			'', '', '', '',
			'', '',
			'', '', '',
		)),
		('\x00', '0', None, (
			'None', '72', '64', 'None', 'None',
			'None', 'None',
			'None', 'None',
			'None', 'None', 'None', 'None',
			'null', 'None',
			'None', 'None', 'None',
		)),
		(':', '"', '"', (
			'"', '72', '64', '"', '"',
			'"', '"',
			'"', '"',
			'"', '"', '"', '"',
			'"', '"',
			'"', '"', '"',
		)),
		(':', '"', {'time': 'never', 'float32': '"0"'}, (
			'None', '72', '64', 'None', 'None',
			'None', 'None',
			'None', 'None',
			'"0"', 'None', 'None', 'None',
			'null', 'None',
			'None', 'never', 'None',
		)),
	):
		with status("Checking with sep=%r, q=%r, none_as=%r" % (sep, q, none_as,)):
			exp = subjobs.build('csvexport', filename='test.csv', separator=sep, source=ds, quote_fields=q, none_as=none_as, lazy_quotes=False)
			with exp.open('test.csv', 'r', encoding='utf-8') as fh:
				def expect(*a):
					want = sep.join(q + v.replace(q, q + q) + q for v in a) + '\n'
					got = next(fh)
					assert want == got, 'wanted %r, got %r from %s (export of %s)' % (want, got, exp, ds,)
				expect(*sorted(todo))
				expect(
					'a', '4294967295', '18364758544493064720', 'True', 'hello',
					'(42+0j)', '(1e+100+1e-17j)',
					'2020-06-23', '2020-06-23 12:13:14',
					'1.0', '-inf', '-10', '-20',
					'{"json": true}', '1203552815971897489538799',
					'...' if PY2 else '(1+2j)', '12:13:14', 'bl\xe5',
				)
				expect(
					'b', '0', '0', 'False', 'bye',
					'(2-3j)', '(-7+0j)',
					'1868-01-03', '1868-01-03 13:14:05',
					'inf', 'nan', '0', '0',
					'[false, null]', '42.18',
					'...' if PY2 else "{'recursion': {...}}", '13:14:05', 'bl\xe4',
				)
				expect(*last_line)



from datetime import date
from .printer import prt

description = "Job options."


def main(urd):
	prt.source(__file__)
	prt()
	prt.header('Examples options')
	prt('Run a job that accepts a variety of options:')
	job = urd.build(
		'example_options',
		day1=date(2021, 10, 14),
		day2=date(2021, 10, 14),
		something=3.14,
		text='hello',
		number=42,
		seq=[1,2,3,4],
		cplx=dict(a=dict(b='c')),
	)
	prt('The job will print its options to stdout.')
	prt()
	prt('See the options as the job sees them by running')
	prt.command('ax job -O %s' % (job,))
	prt('The "ax job" command will also show the job\'s options')
	prt.command('ax job %s' % (job,))
	prt('See everything using')
	prt.command('ax job -o %s' % (job,))

	prt()
	prt.header('Short and long version (see code)')
	prt('Note that')
	urd.build('example_options', text='world')
	prt('is the shorter version of')
	urd.build('example_options', options=dict(text='world'))
	prt('which can be used to avoid name clashes between jobs,')
	prt('options, and datasets.')



from os.path import dirname, join
from .printer import prt

description = "Tutorial: Import several files into a dataset chain"

filenames = ('data.csv', 'data2.csv', 'data3.csv')
# Again, files are stored in same directory as this python file.
# In a real project, set "input directory" in config file instead!
filenames = (join(dirname(__file__), x) for x in filenames)


def main(urd):
	prt.source(__file__)
	prt()
	prt('A loop importing three files, creating a chained dataset of all data')
	prt()

	for filename in filenames:
		prt.plain(filename)
		imp = urd.build(
			'csvimport',
			filename=filename,
			previous=urd.joblist.get('csvimport'),
		)
		imp = urd.build(
			'dataset_type',
			source=imp,
			previous=urd.joblist.get('dataset_type'),
			column2type=dict(
				Date='date:%Y-%m-%d',
				String='unicode:utf-8',
				Int='number',
				Float='float64',
			),
		)
		imp = urd.build(
			'dataset_sort',
			source=imp,
			sort_columns='Date',
			previous=urd.joblist.get('dataset_sort'),
		)
		imp = urd.build(
			'dataset_hashpart',
			source=imp,
			hashlabel='String',
			previous=urd.joblist.get('dataset_hashpart'),
		)

	prt()
	prt('''
		Each job has a dataset input parameter called previous.  We
		assign this input parameter to (a reference) of the job that
		processed the previous file in the list.  This creates a
		dataset chain.

		The point of chaining is that it allows datasets to grow in
		number of rows, with linear time complexity, without changing
		any existing processing state.  This means that we can add
		new rows to a dataset, but we always have access to all previous
		imports of the data.  Nothing is modified, data is just appended.
	''')

	prt()
	with prt.header('INSPECTING DATASET CHAINS'):
		prt('  The flags -s, -S, and -c to "ax ds" are useful when looking')
		prt('  at chained datasets')
		prt.command('ax ds -s %s' % (imp,))
		prt.command('ax ds -S %s' % (imp,))
		prt.command('ax ds -c %s' % (imp,))
		prt('As always, see info and more options using')
		prt.command('ax ds --help')

	prt()
	with prt.header('PRINTING A DATASET CHAIN'):
		prt('This is a small example, so we can print all data using')
		prt.command('ax cat -c %s' % (imp,))

	prt()
	with prt.header('GREPPING'):
		prt('''
			Here's an example of grepping the string "bb" in the
			"String" column, but show "Date", "Float", and "Int"
			columns only.
		''')
		prt.command('ax grep bb -c -g String %s Date Float Int' % (imp,))



############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2019-2021 Carl Drougge                       #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division

from random import randint
from collections import OrderedDict, defaultdict
from itertools import combinations
from copy import deepcopy

from accelerator.compat import iteritems

from accelerator import setupfile
from accelerator import deptree
from accelerator.extras import job_params

def find_possible_jobs(db, methods, job):
	method = job['method']
	params = {method: job['params'][method]}
	optset = methods.params2optset(params)
	if not optset:
		return {}
	def inner():
		for uid, job in db.match_exact([(method, 0, optset,)]):
			yield job.id, ()
			return # no depjobs is enough - stop
		for remset in combinations(optset, remcount):
			for uid, job in db.match_complex([(method, 0, optset - set(remset),)]):
				yield job.id, remset
	res = {}
	remcount = 0
	while not res:
		remcount += 1
		for jobid, remset in inner():
			remset = tuple(s.split()[1] for s in remset)
			res[jobid] = remset
		if remcount == len(optset):
			break
	return dict(_job_candidates_options(res))

def _job_candidates_options(candidates):
	for jobid, remset in iteritems(candidates):
		setup = job_params(jobid)
		optdiff = defaultdict(dict)
		for thing in remset:
			section, name = thing.split('-', 1)
			optdiff[section][name] = setup[section][name]
		yield jobid, optdiff

def initialise_jobs(setup, target_WorkSpace, DataBase, Methods, verbose=False):

	# create a DepTree object used to track options and make status
	DepTree = deptree.DepTree(Methods, setup)

	if not setup.get('force_build'):
		# compare database to deptree
		reqlist = DepTree.get_reqlist()
		for uid, job in DataBase.match_exact(reqlist):
			DepTree.set_link(uid, job)
	DepTree.propagate_make()
	why_build = setup.get('why_build')
	if why_build:
		orig_params = deepcopy(DepTree.params)
	DepTree.fill_in_default_options()

	# get list of jobs in execution order
	joblist = DepTree.get_sorted_joblist()
	newjoblist = [x for x in joblist if x['make']]
	num_new_jobs = len(newjoblist)

	if why_build == True or (why_build and num_new_jobs):
		res = OrderedDict()
		DepTree.params = orig_params
		joblist = DepTree.get_sorted_joblist()
		for job in joblist:
			if job['make']:
				res[job['method']] = find_possible_jobs(DataBase, Methods, job)
			else:
				res[job['method']] = {job['link']: {}}
		return [], {'why_build': res}

	if num_new_jobs:
		new_jobid_list = target_WorkSpace.allocate_jobs(num_new_jobs)
		# insert new jobids
		for (x,jid) in zip(newjoblist, new_jobid_list):
			x['link'] = jid
		for data in newjoblist:
			method = Methods.db[data['method']]
			params = data['params'][data['method']]
			new_setup = setupfile.generate(
				caption=setup.caption,
				method=data['method'],
				options=params['options'],
				datasets=params['datasets'],
				jobs=params['jobs'],
				package=method['package'],
				description=Methods.descriptions[data['method']],
			)
			new_setup.hash = Methods.hash[data['method']][0]
			new_setup.seed = randint(0, 2**63 - 1)
			new_setup.jobid = data['link']
			new_setup.slices = target_WorkSpace.slices
			typing = Methods.typing[data['method']]
			if typing:
				new_setup['_typing'] = typing
			setupfile.save_setup(data['link'], new_setup)
	else:
		new_jobid_list = []

	res = {j['method']: {k: v for k, v in j.items() if k in ('link', 'make', 'total_time')} for j in joblist}
	return new_jobid_list, {'jobs': res}



############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2018-2020 Carl Drougge                       #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

description = r'''
Take a dataset and make a checksum of one or more columns.

This is a debugging aid, so you can compare datasets across machines,
even with different slicing.

If you set options.sort=False it's faster, but your datasets have to have
the same slicing and order to compare equal.

options.columns defaults to all columns in the source dataset.

Note that this uses about 64 bytes of RAM per line, so you can't sum huge
datasets. (So one GB per 20M lines or so.)
'''

from hashlib import md5
from itertools import chain
from json import JSONEncoder
from heapq import merge

from accelerator.extras import DotDict
from accelerator.compat import PY2

options = dict(
	columns      = set(),
	sort         = True,
)

datasets = ('source',)

if PY2:
	bytesrepr = repr
else:
	def bytesrepr(v):
		return repr(v).encode('utf-8')

def bytesstr(v):
	return v.encode('utf-8')

def bytesstr_none(v):
	return b'None\xff' if v is None else v.encode('utf-8')

def self_none(v):
	return b'None\xff' if v is None else v

jsonenc = JSONEncoder(
	sort_keys=True,
	ensure_ascii=False,
	check_circular=False,
	separators=(',', ':'),
).encode
def sortdicts(v):
	return jsonenc(v).encode('utf-8')

def prepare():
	columns = sorted(options.columns or datasets.source.columns)
	translators = {}
	# Special casing the string types here is just a performance
	# optimisation. (Which changes the hash, but not how good it is.)
	# Special casing json is needed because the iteration order of
	# dicts is not fixed (depending on python version).
	# Same with pickle, but worse (many picklable values will break this).
	for n in columns:
		col = datasets.source.columns[n]
		if col.type == 'bytes' or (col.type == 'ascii' and PY2):
			# doesn't need any encoding, but might need None-handling.
			if col.none_support:
				translators[n] = self_none
		elif col.type in ('unicode', 'ascii'):
			if col.none_support:
				translators[n] = bytesstr_none
			else:
				translators[n] = bytesstr
		elif col.type == 'json':
			translators[n] = sortdicts
		elif col.type == 'pickle':
			translators[n] = sortdicts
			print('WARNING: Column %s is pickle, may not work' % (n,))
		else:
			translators[n] = bytesrepr
	return columns, translators

def analysis(sliceno, prepare_res):
	columns, translators = prepare_res
	src = datasets.source.iterate(sliceno, columns, translators=translators)
	res = []
	for line in src:
		m = md5()
		for item in line:
			m.update(item)
			m.update(b'\0') # Some separator is better than nothing.
		res.append(m.digest())
	if options.sort:
		res.sort()
	return res

def synthesis(prepare_res, analysis_res):
	if options.sort:
		all = merge(*analysis_res)
	else:
		all = chain.from_iterable(analysis_res)
	res = md5(b''.join(all)).hexdigest()
	print("%s: %s" % (datasets.source, res,))
	return DotDict(sum=int(res, 16), sort=options.sort, columns=prepare_res[0], source=datasets.source)



############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2019-2022 Carl Drougge                       #
# Modifications copyright (c) 2019-2021 Anders Berkeman                    #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import division, print_function

import sys
import locale
from datetime import datetime, time, date
from math import ceil, floor, log10, isinf, isnan

from accelerator.compat import ArgumentParser
from accelerator.compat import terminal_size
from .parser import name2ds, name2job
from accelerator.colourwrapper import colour
from accelerator.error import NoSuchWhateverError
from accelerator.extras import quote

MINMAXWIDTH = 13 # minimum number of characters reserved for min/max values
COLUMNS, LINES = terminal_size()


def colwidth(rows):
	# find max string len per column
	return [max(len(s) for s in col) for col in zip(*rows)]

def printcolwise(data, template, printfunc, minrows=8, indent=4):
	if not data:
		return
	cols = (COLUMNS - indent) // (len(template.format(*printfunc(data[0]))) + 2)
	n = int(ceil(len(data) / cols))
	n = max(n, minrows)
	for r in range(n):
		v = data[r::n]
		if v:
			print(' ' * indent + '  '.join(template.format(*printfunc(x)) for x in v))

def main(argv, cfg):
	usage = "%(prog)s [options] ds [ds [...]]"
	parser = ArgumentParser(prog=argv.pop(0), usage=usage)
	parser.add_argument('-c', '--chain', action='store_true', help='list all datasets in a chain')
	parser.add_argument('-C', '--non-empty-chain', action='store_true', help='list all non-empty datasets in a chain')
	parser.add_argument('-l', '--list', action='store_true', help='list all datasets in a job with number of rows')
	parser.add_argument('-L', '--chainedlist', action='store_true', help='list all datasets in a job with number of chained rows')
	parser.add_argument('-m', '--suppress-minmax', action='store_true', help='do not print min/max column values')
	parser.add_argument('-n', '--suppress-columns', action='store_true', help='do not print columns')
	parser.add_argument('-q', '--suppress-errors', action='store_true', help='silently ignores bad input datasets/jobids')
	parser.add_argument('-s', '--slices', action='store_true', help='list relative number of lines per slice in sorted order')
	parser.add_argument('-S', '--chainedslices', action='store_true', help='same as -s but for full chain')
	parser.add_argument('-w', '--location', action='store_true', help='show where (ds/filename) each column is stored')
	parser.add_argument("dataset", nargs='+', help='the job part of the dataset name can be specified in the same ways as for "ax job". you can use ds~ or ds~N to follow the chain N steps backwards, or ^ to follow .parent. this requires specifying the ds-name, so wd-1~ will not do this, but wd-1/default~ will.')
	args = parser.parse_intermixed_args(argv)
	args.chain = args.chain or args.non_empty_chain

	def finish(badinput):
		if badinput and not args.suppress_errors:
			print('Error, failed to resolve datasets:', file=sys.stderr)
			for n, e in badinput:
				print('    %r: %s' % (n, e,), file=sys.stderr)
			exit(1)
		exit()

	badinput = []

	if args.list or args.chainedlist:
		for n in args.dataset:
			try:
				try:
					dsvec = name2ds(cfg, n).job.datasets
				except NoSuchWhateverError:
					dsvec = name2job(cfg, n).datasets
			except Exception as e:
				badinput.append((n, e))
				dsvec = None
			if dsvec:
				print('%s' % (dsvec[0].job,))
				v = []
				for ds in dsvec:
					if args.chainedlist:
						lines = sum(sum(x.lines) for x in ds.chain())
					else:
						lines = sum(ds.lines)
					v.append((ds.quoted, '{:n}'.format(lines)))
				len_n, len_l = colwidth(v)
				template = "{0:%d}  ({1:>%d})" % (len_n, len_l)
				for name, numlines in sorted(v):
					print('    ' + template.format(name, numlines))
		finish(badinput)

	for n in args.dataset:
		try:
			ds = name2ds(cfg, n)
		except NoSuchWhateverError as e:
			badinput.append((n, e))
			continue

		print(ds.quoted)
		if ds.parent:
			if isinstance(ds.parent, tuple):
				print("    Parents:")
				max_n = max(len(x.quoted) for x in ds.parent)
				template = "{1:%d}" % (max_n,)
				data = tuple((None, x.quoted) for ix, x in enumerate(ds.parent))
				data = sorted(data, key = lambda x: x[1])
				printcolwise(data, template, lambda x: x, minrows=8, indent=8)
			else:
				print("    Parent:", ds.parent.quoted)
		print("    Method:", quote(ds.job.method))
		if ds.filename:
			print("    Filename:", quote(ds.filename))
		if ds.previous:
			print("    Previous:", ds.previous.quoted)
		if ds.hashlabel is not None:
			print("    Hashlabel:", quote(ds.hashlabel))

		def prettyminmax(minval, maxval):
			if args.suppress_minmax:
				return ''
			s = '[%%%ds, %%%ds]' % (MINMAXWIDTH, MINMAXWIDTH)
			if minval is None:
				return ''
			elif isinstance(minval, float):
				def intdigits(x):
					if isinf(x) or isnan(x):
						return 3
					return min(MINMAXWIDTH - 2, floor(log10(abs(x)) + 1)) if x else (MINMAXWIDTH - 2)//2
				ints = max(intdigits(minval), intdigits(maxval))
				if ints > 0:
					format = "%% %d.%df" % (ints, MINMAXWIDTH - ints - 2)
				elif ints < -4:
					format = "%% .%de" % (MINMAXWIDTH - 7,)
				else:
					format = "%% .%df" % (MINMAXWIDTH - 3,)
				def format_or_int(v):
					try:
						i = int(v)
						if v == i:
							return i
					except (OverflowError, ValueError):
						pass
					return locale.format_string(format, v)
				return s % (format_or_int(minval), format_or_int(maxval))
			elif isinstance(minval, int):
				return s % (minval, maxval)
			elif isinstance(minval, (date, time, datetime)):
				return s % (minval, maxval)
			else:
				return s % (minval, maxval)

		if not args.suppress_columns:
			print("    Columns:")
			name2typ = {n: c.type + '+None' if c.none_support else c.type for n, c in ds.columns.items()}
			len_n, len_t = colwidth((quote(n), name2typ[n]) for n, c in ds.columns.items())
			if args.location:
				len_l = max(len(quote(c.location)) for c in ds.columns.values())
				len_c = max(len(c.compression) for c in ds.columns.values())
				template = '        {2} {0:%d}  {1:%d}  {4:%d} {5:%d}  {3}' % (len_n, len_t, len_l, len_c,)
			else:
				template = '        {2} {0:%d}  {1:%d}  {3}' % (len_n, len_t,)
			chain = False
			if args.chainedslices or args.chain:
				chain = ds.chain()
			for n, c in sorted(ds.columns.items()):
				if chain:
					minval, maxval = chain.min(n), chain.max(n)
				else:
					minval, maxval = c.min, c.max
				hashdot = colour("*", "ds/highlight") if n == ds.hashlabel else " "
				print(template.format(quote(n), name2typ[n], hashdot, prettyminmax(minval, maxval), quote(c.location), c.compression).rstrip())
			print("    {0:n} columns".format(len(ds.columns)))
		print("    {0:n} lines".format(sum(ds.lines)))

		if ds.previous or args.chain:
			chain = ds.chain()
			if args.non_empty_chain:
				print("    Full chain length {0:n}, from {1} to {2}".format(len(chain), chain[0], chain[-1]))
				chain = [ds for ds in chain if sum(ds.lines)]
				print("    Filtered chain length {0:n}".format(len(chain)))
			if chain:
				if not args.non_empty_chain:
					print("    Chain length {0:n}, from {1} to {2}".format(len(chain), chain[0], chain[-1]))
				if args.chain:
					data = tuple((ix, "%s/%s" % (x.job, x.name), "{:n}".format(sum(x.lines))) for ix, x in enumerate(chain))
					max_n, max_l = colwidth(x[1:] for x in data)
					template = "{0:3}: {1:%d} ({2:>%d})" % (max_n, max_l)
					printcolwise(data, template, lambda x: (x[0], x[1], x[2]), minrows=8, indent=8)

		if args.slices or args.chainedslices:
			if args.chainedslices and ds.previous:
				data = ((ix, '{:n}'.format(sum(x)), sum(x)) for ix, x in enumerate(zip(*(x.lines for x in ds.chain()))))
				print('    Balance, lines per slice, full chain:')
			else:
				data = ((ix, '{:n}'.format(x), x) for ix, x in enumerate(ds.lines))
				if ds.previous:
					print('    Balance, lines per slice, tip dataset:')
				else:
					print('    Balance, lines per slice:')
			data = sorted(data, key=lambda x: -x[2])
			s = sum(x[2] for x in data)
			len_n = max(len(x[1]) for x in data)
			template = "{0:3}: {1!s}%% ({2:>%d})" % (len_n,)
			printcolwise(data, template, lambda x: (x[0], locale.format_string("%6.2f", (100 * x[2] / (s or 1e20))), x[1]), minrows=8, indent=8)
			print("    Max to average ratio: " + locale.format_string("%2.3f", (max(x[2] for x in data) / ((s or 1e20) / len(data)),) ))

		if ds.previous:
			print("    {0:n} total lines in chain".format(sum(sum(ds.lines) for ds in chain)))

	finish(badinput)



############################################################################
#                                                                          #
# Copyright (c) 2022 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

description = r'''
Test the various :parsed types on DatasetWriter
'''

def synthesis(job):
	inf = float('inf')
	for none_support in (False, True):
		for t, parse_values, values in (
			('int64', ['-1', '0', '999999999999999999'], [-1, 0, 999999999999999999]),
			('int32', ['-1', '0', '999999999'], [-1, 0, 999999999]),
			('bits64', ['0', '1', '9999999999999999999'], [0, 1, 9999999999999999999]),
			('bits32', ['0', '1', '3999999999'], [0, 1, 3999999999]),
			('float64', ['1', '-2', '3.3'], [1.0, -2.0, 3.3]),
			('float32', ['inf', '0.1', '-inf'], [inf, 0.10000000149011612, -inf]),
			('complex64', ['1', '2j', '(-3.3)', '4-5j'], [1+0j, 0+2j, -3.3+0j, 4-5j]),
			('complex32', ['infj', 'inf-infj', 'inf+0.5j'], [complex(0, inf), complex(inf, -inf), complex(inf, 0.5)]),
			('number', ['-1', '2.2', '99999999999999999999'], [-1, 2.2, 99999999999999999999]),
			('json', ['1', '["2"]', '{"3": 3.3}'], [1, ['2'], {'3': 3.3}]),
		):
			if none_support and t.startswith('bits'):
				# no none_support in the bits types
				continue
			if t == 'json':
				bad_value = object
			else:
				bad_value = 'not valid'
			name = '%s %s none_support' % (t, 'with' if none_support else 'without',)
			dw = job.datasetwriter(name=name, allow_missing_slices=True)
			dw.add('data', 'parsed:' + t, none_support=none_support)
			dw.set_slice(0)
			for v in values:
				dw.write(v)
			dw.set_slice(1)
			for v in parse_values:
				assert isinstance(v, str), 'oops: %r' % (v,)
				dw.write(v)
			try:
				dw.write(bad_value)
				raise Exception("parsed:%s accepted %r as a value" % (t, bad_value,))
			except (ValueError, TypeError):
				pass
			# json will of course accept None even without none_support
			apparent_none_support = (none_support or t == 'json')
			dw.set_slice(2)
			try:
				dw.write(None)
				if not apparent_none_support:
					raise Exception('parsed:%s accepted None without none_support' % (t,))
			except (ValueError, TypeError):
				if apparent_none_support:
					raise Exception('parsed:%s did not accept None despite none_support' % (t,))
			ds = dw.finish()
			for sliceno, desc in enumerate(("normal values", "parseable values",)):
				got = list(ds.iterate(sliceno, 'data'))
				assert got == values, "parsed:%s (%s) %s gave %r, wanted %r" % (t, ds.quoted, desc, got, values,)
			if apparent_none_support:
				got = list(ds.iterate(2, 'data'))
				assert got == [None], "parsed:%s (%s) gave %r, wanted [None]" % (t, ds.quoted, got,)
			dw = job.datasetwriter(name=name + ' with default', allow_missing_slices=True)
			default = None if none_support else 42
			dw.add('data', 'parsed:' + t, none_support=none_support, default=default)
			dw.set_slice(0)
			dw.write('1')
			dw.write(bad_value)
			ds = dw.finish()
			got = list(ds.iterate(0, 'data'))
			assert got == [1, default], "parsed:%s with default=%s (%s) gave %r, wanted [1, %s]" % (t, default, ds.quoted, got, default)



############################################################################
#                                                                          #
# Copyright (c) 2020-2021 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

import sys
import os

from accelerator.compat import ArgumentParser, url_quote
from accelerator.unixhttp import call
from accelerator.setupfile import load_setup
from accelerator.build import fmttime
from accelerator.extras import DotDict
from accelerator.job import Job

def job_data(known, jid, full_path=False):
	if jid in known:
		data = known[jid]
	else:
		data = DotDict(method='???', totaltime=None, current=None)
		try:
			setup = load_setup(jid)
			data.method = setup.method
			if 'exectime' in setup:
				data.totaltime = setup.exectime.total
		except Exception:
			pass
	if isinstance(data.totaltime, (float, int)):
		data.totaltime = fmttime(data.totaltime)
	if data.totaltime is None:
		data.klass = 'unfinished'
	elif data.current:
		data.klass = 'current'
	else:
		data.klass = 'old'
	if not data.get('path'):
		if full_path:
			data.path = Job(jid).path
		else:
			data.path = jid
	return data

def show_job(args, known, jid, as_latest=False):
	data = job_data(known, jid, args.full_path)
	path = data.path
	if as_latest:
		path = path.rsplit('-', 1)[0] + '-LATEST'
	print('\t'.join((path, data.klass, data.method, data.totaltime or '')))

def workdir_jids(cfg, name):
	jidlist = []
	for jid in os.listdir(cfg.workdirs[name]):
		if '-' in jid:
			wd, num = jid.rsplit('-', 1)
			if wd == name and num.isdigit():
				jidlist.append(int(num))
	jidlist.sort()
	return ['%s-%s' % (name, jid,) for jid in jidlist]

def main(argv, cfg):
	usage = "%(prog)s [-p] [-a | [workdir [workdir [...]]]"
	parser = ArgumentParser(usage=usage, prog=argv.pop(0))
	parser.add_argument('-a', '--all', action='store_true', help="list all workdirs")
	parser.add_argument('-p', '--full-path', action='store_true', help="show full path")
	parser.add_argument('workdirs', nargs='*', default=[])
	args = parser.parse_args(argv)

	if args.all:
		args.workdirs.extend(sorted(set(cfg.workdirs) - set(args.workdirs)))

	if not args.workdirs:
		template = '%%-%ds  %%s' % (max(len(wd) for wd in cfg.workdirs))
		for wd, path in sorted(cfg.workdirs.items()):
			print(template % (wd, path,))
		return

	for name in args.workdirs:
		if name not in cfg.workdirs:
			print("No such workdir:", name, file=sys.stderr)
			continue
		known = call(cfg.url + '/workdir/' + url_quote(name))
		for jid in workdir_jids(cfg, name):
			show_job(args, known, jid)

		try:
			latest = os.readlink(os.path.join(cfg.workdirs[name], name + '-LATEST'))
		except OSError:
			latest = None
		if latest:
			show_job(args, known, jid, True)



from os.path import dirname, join
from .printer import prt

description = "Jobs: equivalent_hashes howto"


def main(urd):
	prt.source(__file__)
	prt()
	prt('This example shows how to create an "equivalent hash", i.e.')
	prt('how to modify a method without causing a build script to')
	prt('re-build it.')
	job = urd.build('example_equivalenthashes')

	thehash = job.params.hash
	prt()
	prt('Take the hash from the built job:')
	prt.output('"%s"' % (thehash,))
	prt('and add it to the method like this')
	prt.output('"equivalent_hashes = {\'whatever\': (\'%s\',)}"' % (thehash,))

	prt()
	prt('Also do some other change to the method while at it to make it different.')
	prt('Re-run and the daemon output will be something like')
	prt.output('''
		==========================================================================================
		WARNING: a_example_equivalenthashes has equivalent_hashes, but missing verifier <verifier>
		==========================================================================================
	''')
	prt('Now, let <verifier> replace the string \'whatever\' in the method.')
	prt('Done, the new version of the method is now considered equivalent')
	prt('to the old one.')
	prt()
	prt('Method source file is here')
	prt.plain(join(dirname(__file__), 'a_%s.py' % (job.method,)))



def prepare(job):
	dw1 = job.datasetwriter(name='first')
	dw2 = job.datasetwriter(name='second')
	dw3 = job.datasetwriter(name='third')
	dw1.add('col1', 'int64')
	dw2.add('col1', 'json')
	dw3.add('col1', 'number')
	dw3.add('col2', 'ascii')
	dw3.add('col3', 'bool')
	return dw1, dw2, dw3


def analysis(sliceno, prepare_res):
	dw1, dw2, dw3 = prepare_res
	dw1.write(sliceno)
	dw2.write({'sliceno': sliceno})
	dw3.write(sliceno, str(sliceno), sliceno % 2 == 0)



############################################################################
#                                                                          #
# Copyright (c) 2021 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

description = r'''
Verify that lazy_quotes quotes the expected values in csvexport with
various quotes and separators.
'''

from accelerator import subjobs
from accelerator.compat import PY3

def test(job, prefix, none_support):
	expect = [[], [], []]
	def write(sliceno, a, b, c):
		w(a, b, c)
		if isinstance(c, bytes):
			if PY3:
				c = c.decode('utf-8', 'backslashreplace')
		else:
			c = repr(c)
		expect[sliceno].append((str(a), repr(b), c,))
	dw = job.datasetwriter(name=prefix + 'a')
	dw.add('a', 'ascii', none_support=none_support)
	dw.add('b', 'int32', none_support=none_support)
	dw.add('c', 'float64', none_support=none_support)
	w = dw.get_split_write()
	write(0, 'hello', 42, float('-inf'))
	write(1, "'world'", -1, 5e-33)
	if none_support:
		write(2, None, None, None)
	a = dw.finish()
	dw = job.datasetwriter(name=prefix + 'b', previous=a)
	dw.add('a', 'bool', none_support=none_support)
	dw.add('b', 'number', none_support=none_support)
	dw.add('c', 'bytes', none_support=none_support)
	w = dw.get_split_write()
	write(0, False, float('nan'), b"'ascii'")
	write(1, True, 1.2e+22, b'\xff')
	if none_support:
		write(2, None, None, None)
	b = dw.finish()
	expect = [('a', 'b', 'c',)] + expect[0] + expect[1] + expect[2] + [()]
	for lazy_quotes in (False, True,):
		for q in ('', '42', '-', 'f', 'hello',):
			for sep in ('+', 'a', '',):
				verify(b, lazy_quotes, q, sep, expect)
	if none_support:
		expect = [none2nah(line) for line in expect]
		verify(b, True, '"', 'ah', expect, none_as='nah')
		verify(b, True, "'", '"', expect, none_as='nah')
		verify(b, True, 'a', ',', expect, none_as='nah')

def verify(source, lazy_quotes, q, sep, expect, **kw):
	j = subjobs.build(
		'csvexport',
		chain_source=True,
		source=source,
		lazy_quotes=lazy_quotes,
		quote_fields=q,
		separator=sep,
		**kw
	)
	with j.open('result.csv', 'r' if PY3 else 'rb') as fh:
		got = fh.read()
	if lazy_quotes and sep:
		quote_func = make_lazy(sep, q)
	else:
		quote_func = lambda v: q + v.replace(q, q + q) + q
	want = '\n'.join(sep.join(map(quote_func, line)) for line in expect)
	if want != got:
		print('Unhappy with %s:' % (j.filename('result.csv'),))
		print()
		print('Expected:')
		print(want)
		print('Got:')
		print(got)
		raise Exception('csvexport failed with quote_fields=%r, separator=%r, lazy_quotes=%r' % (q, sep, lazy_quotes,))

def make_lazy(sep, q):
	if q == '"':
		def quote(v):
			if v.startswith(q) or v.endswith(q) or sep in v or v.startswith("'") or v.endswith("'"):
				return q + v.replace(q, q + q) + q
			else:
				return v
	else:
		def quote(v):
			if v.startswith(q) or v.endswith(q) or sep in v:
				return q + v.replace(q, q + q) + q
			else:
				return v
	return quote

def none2nah(line):
	return ['nah' if v == 'None' else v for v in line]

def synthesis(job):
	test(job, '', False)
	test(job, 'None ', True)



############################################################################
#                                                                          #
# Copyright (c) 2019 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test dataset_sort as a chain, across a chain and as a chain merging
only two datasets of the original chain.
'''

from accelerator import subjobs
from accelerator.dataset import Dataset, DatasetWriter

def synthesis():
	dw_a = DatasetWriter(name='a', columns={'num': 'int32'})
	dw_b = DatasetWriter(name='b', columns={'num': 'int32'}, previous=dw_a)
	dw_c = DatasetWriter(name='c', columns={'num': 'int32'}, previous=dw_b)
	w = dw_a.get_split_write()
	w(3)
	w(2)
	w = dw_b.get_split_write()
	w(2)
	w(1)
	w = dw_c.get_split_write()
	w(0)
	a = dw_a.finish()
	b = dw_b.finish()
	c = dw_c.finish()

	opts = dict(
		sort_columns='num',
		sort_across_slices=True,
	)

	# sort as a chain
	jid = subjobs.build('dataset_sort', options=opts, datasets=dict(source=a, previous=None))
	assert list(Dataset(jid).iterate(None, 'num')) == [2, 3]
	sorted_a = jid
	jid = subjobs.build('dataset_sort', options=opts, datasets=dict(source=b, previous=jid))
	assert list(Dataset(jid).iterate_chain(None, 'num')) == [2, 3, 1, 2]
	jid = subjobs.build('dataset_sort', options=opts, datasets=dict(source=c, previous=jid))
	assert list(Dataset(jid).iterate_chain(None, 'num')) == [2, 3, 1, 2, 0]

	# sort all as a single dataset
	jid = subjobs.build('dataset_sort', options=opts, datasets=dict(source=c, previous=None))
	assert list(Dataset(jid).iterate_chain(None, 'num')) == [0, 1, 2, 2, 3]

	# merge b and c but not a
	jid = subjobs.build('dataset_sort', options=opts, datasets=dict(source=c, previous=sorted_a))
	# test with new style job.dataset
	assert list(jid.dataset().iterate(None, 'num')) == [0, 1, 2]
	assert list(jid.dataset().iterate_chain(None, 'num')) == [2, 3, 0, 1, 2]



from .printer import prt

description = "Jobs: prepare, analysis, and synthesis + return values"


def main(urd):
	prt.source(__file__)
	prt()
	prt('Run a job that returns (and prints) something in all of')
	prt('prepare, analysis, and synthesis.')
	job = urd.build('example_returninprepanasyn')

	prt()
	prt('Job output:')
	prt.output(job.output())

	prt()
	prt('Job return value:')
	prt.output(job.load())



############################################################################
#                                                                          #
# Copyright (c) 2019 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
This method is called to test that urd.build finds the correct target for
keywords, and also that it does not accept ambiguous keywords.
'''

options = {'foo': 'foo', 'a': 'a'}
datasets = ('foo', 'b')
jobs = ('foo', 'c')

def synthesis():
	return (options, datasets, jobs)



from accelerator.job import JobWithFile

description = 'JobWithFile, read normal and sliced files'

options = dict(firstfile=JobWithFile, secondfile=JobWithFile)


def analysis(sliceno):
	print(options.firstfile.load(sliceno=sliceno))


def synthesis():
	print(options.secondfile.load())



############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2019-2022 Carl Drougge                       #
# Modifications copyright (c) 2020 Anders Berkeman                         #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

# grep in a dataset(chain)

from __future__ import division, print_function

import sys
import re
import os
from itertools import chain, repeat, cycle
from collections import deque, defaultdict
from argparse import RawTextHelpFormatter, Action, SUPPRESS, ArgumentError
from multiprocessing import Lock
import json
import datetime
import operator
import signal

from accelerator.compat import ArgumentParser
from accelerator.compat import unicode, izip, PY2
from accelerator.compat import monotonic
from accelerator.compat import QueueEmpty
from accelerator.colourwrapper import colour
from .parser import name2ds
from accelerator.error import NoSuchWhateverError
from accelerator.extras import DotDict
from accelerator import g
from accelerator import mp


# os.write can't be trusted to write everything
def write(fd, data):
	while data:
		data = data[os.write(fd, data):]


def main(argv, cfg):
	# -C overrides -A and -B (which in turn override -C)
	class ContextAction(Action):
		def __call__(self, parser, namespace, values, option_string=None):
			namespace.before_context = namespace.after_context = values

	# Default values for tab_length
	tab_length = DotDict(tab_len=8, field_len=16, min_len=2)

	class TabLengthAction(Action):
		def __call__(self, parser, namespace, values, option_string=None):
			names = {}
			for suffix in ('_length', 'length', '_len', 'len', ''):
				for name in ('tab', 'field', 'min'):
					names[name + suffix] = name + '_len'
			min_value = dict(tab_len=1, field_len=0, min_len=1)
			order = ['tab_len', 'field_len', 'min_len']
			unnamed = cycle(order)
			values = re.split(r'[/, ]', values)
			if len(values) == 1 and values[0] and '=' not in values[0]:
				# only a single number provided: skip defaults and be traditional
				tab_length.field_len = 0
				tab_length.min_len = 1
			for value in values:
				if not value:
					continue # so you can do -T/ or similar to just activate it
				if '=' in value:
					name, value = value.rsplit('=', 1)
					name = name.lower()
				else:
					name = next(unnamed)
				if name not in names:
					raise ArgumentError(self, 'unknown field %r' % (name,))
				name = names[name]
				try:
					value = int(value)
				except ValueError:
					raise ArgumentError(self, 'invalid int value for %s: %r' % (name, value,))
				if value < min_value[name] or value > 9999:
					raise ArgumentError(self, 'invalid value for %s: %d' % (name, value,))
				tab_length[name] = value
			# -T overrides -t
			namespace.separator = None
			namespace.tab_length = True

	class SeparatorAction(Action):
		def __call__(self, parser, namespace, values, option_string=None):
			# -t overrides -T
			namespace.tab_length = False
			namespace.separator = values

	parser = ArgumentParser(
		usage="%(prog)s [options] [-e] pattern [...] [-d] ds [...] [[-n] column [...]]",
		description="""positional arguments:
  pattern               (-e, --regexp)
  dataset               (-d, --dataset) can be specified as for "ax ds"
  columns               (-n, --column)""",
		prog=argv.pop(0),
		formatter_class=RawTextHelpFormatter,
	)
	parser.add_argument('-c', '--chain',        action='store_true', help="follow dataset chains", )
	parser.add_argument(      '--chain-length', '--cl', metavar='LENGTH', action='append', type=int, help="follow chains at most this many datasets")
	parser.add_argument(      '--stop-ds',      action='append', metavar='DATASET', help="follow chains at most to this dataset\nthese options work like in ds.chain() and you can\nspecify them several times (for several datasets)")
	parser.add_argument(      '--colour', '--color', nargs='?', const='always', choices=['auto', 'never', 'always'], type=str.lower, help="colour matched text. can be auto, never or always", metavar='WHEN', )
	parser.add_argument('-i', '--ignore-case',  action='store_true', help="case insensitive pattern", )
	parser.add_argument('-v', '--invert-match', action='store_true', help="select non-matching lines", )
	parser.add_argument('-o', '--only-matching',action='store_true', help="only print matching part (or columns with -l)", )
	parser.add_argument('-l', '--list-matching',action='store_true', help="only print matching datasets (or slices with -S)\nwhen used with -o, only print matching columns", )
	parser.add_argument('-H', '--headers',      action='store_true', help="print column names before output (and on each change)", )
	parser.add_argument('-O', '--ordered',      action='store_true', help="output in order (one slice at a time)", )
	parser.add_argument('-M', '--allow-missing-columns', action='store_true', help="datasets are allowed to not have (some) columns", )
	parser.add_argument('-g', '--grep',         action='append',     help="grep this column only, can be specified multiple times", metavar='COLUMN')
	parser.add_argument('-s', '--slice',        action='append',     help="grep this slice only, can be specified multiple times",  type=int)
	parser.add_argument('-D', '--show-dataset', action='store_true', help="show dataset on matching lines", )
	parser.add_argument('-S', '--show-sliceno', action='store_true', help="show sliceno on matching lines", )
	parser.add_argument('-L', '--show-lineno',  action='store_true', help="show lineno (per slice) on matching lines", )
	supported_formats = ('csv', 'raw', 'json',)
	parser.add_argument('-f', '--format', default='csv', choices=supported_formats, help="output format, csv (default) / " + ' / '.join(supported_formats[1:]), metavar='FORMAT', )
	parser.add_argument('-t', '--separator', action=SeparatorAction, help="field separator, default tab / tab-like spaces", )
	parser.add_argument('-T', '--tab-length', metavar='LENGTH', action=TabLengthAction,
		help="field alignment, always uses spaces as separator\n" +
		     "specify as many as you like of TABLEN/FIELDLEN/MINLEN\n" +
		     "or as NAME=VALUE (e.g. \"min=3/field=24\")\n" +
		     "TABLEN works like normal tabs\n" +
		     "FIELDLEN sets a longer minimum between fields\n" +
		     "MINLEN sets a minimum len for all separators\n" +
		     "use \"-T/\" to just activate it (sets %d/%d/%d)" % (tab_length.tab_len, tab_length.field_len, tab_length.min_len,)
	)
	parser.add_argument('-B', '--before-context', type=int, default=0, metavar='NUM', help="print NUM lines of leading context", )
	parser.add_argument('-A', '--after-context',  type=int, default=0, metavar='NUM', help="print NUM lines of trailing context", )
	parser.add_argument('-C', '--context',        type=int, default=0, metavar='NUM', action=ContextAction,
		help="print NUM lines of context\n" +
		     "context is only taken from the same slice of the same\n" +
		     "dataset, and may intermix with output from other\n" +
		     "slices. Use -O to avoid that, or -S -L to see it.",
	)
	parser.add_argument('-e', '--regexp',  default=[], action='append', dest='patterns', help=SUPPRESS)
	parser.add_argument('-d', '--dataset', default=[], action='append', dest='datasets', help=SUPPRESS)
	parser.add_argument('-n', '--column',  default=[], action='append', dest='columns', help=SUPPRESS)
	parser.add_argument('words', nargs='*', help=SUPPRESS)
	args = parser.parse_intermixed_args(argv)

	if args.before_context < 0 or args.after_context < 0:
		print('Context must be >= 0', file=sys.stderr)
		return 1

	columns = args.columns

	try:
		args.datasets = [name2ds(cfg, ds) for ds in args.datasets]
	except NoSuchWhateverError as e:
		print(e, file=sys.stderr)
		return 1

	for word in args.words:
		if not args.patterns:
			args.patterns.append(word)
		elif columns and args.datasets:
			columns.append(word)
		else:
			try:
				args.datasets.append(name2ds(cfg, word))
			except NoSuchWhateverError as e:
				if not args.datasets:
					print(e, file=sys.stderr)
					return 1
				columns.append(word)

	if not args.patterns or not args.datasets:
		parser.print_help(file=sys.stderr)
		return 1

	datasets = args.datasets
	patterns = []
	for pattern in args.patterns:
		try:
			patterns.append(re.compile(pattern, re.IGNORECASE if args.ignore_case else 0))
		except re.error as e:
			print("Bad pattern %r:\n%s" % (pattern, e,), file=sys.stderr)
			return 1

	grep_columns = set(args.grep or ())
	if grep_columns == set(columns):
		grep_columns = set()

	if args.slice:
		want_slices = []
		for s in args.slice:
			assert 0 <= s < g.slices, "Slice %d not available" % (s,)
			if s not in want_slices:
				want_slices.append(s)
	else:
		want_slices = list(range(g.slices))

	if len(want_slices) == 1:
		# it will be automatically ordered, so let's not work for it.
		args.ordered = False

	if args.only_matching:
		if args.list_matching:
			args.list_matching = False
			only_matching = 'columns'
		else:
			only_matching = 'part'
	else:
		only_matching = False

	if args.chain:
		datasets = list(chain.from_iterable(
			ds.chain(length=length, stop_ds=stop_ds)
			for ds, length, stop_ds in zip(
				datasets,
				cycle(args.chain_length or [-1]),
				cycle(args.stop_ds or [None]),
			)
		))

	def columns_for_ds(ds, columns=columns):
		if columns:
			return [n for n in columns if n in ds.columns]
		else:
			return sorted(ds.columns)

	if columns or grep_columns:
		if args.allow_missing_columns:
			keep_datasets = []
			for ds in datasets:
				if not columns_for_ds(ds):
					continue
				if grep_columns and not columns_for_ds(ds, grep_columns):
					continue
				keep_datasets.append(ds)
			if not keep_datasets:
				return 0
			datasets = keep_datasets
		else:
			bad = False
			need_cols = set(columns)
			if grep_columns:
				need_cols.update(grep_columns)
			for ds in datasets:
				missing = need_cols - set(ds.columns)
				if missing:
					print('ERROR: %s does not have columns %r' % (ds, missing,), file=sys.stderr)
					bad = True
			if bad:
				return 1

	# For the status reporting, this gives how many lines have been processed
	# when reaching each ds ix, per slice. Ends with an extra fictional ds,
	# i.e. the total number of lines for that slice. And then the same again,
	# to simplify the code in the status shower.
	total_lines_per_slice_at_ds = [[0] * g.slices]
	for ds in datasets:
		total_lines_per_slice_at_ds.append([a + b for a, b in zip(total_lines_per_slice_at_ds[-1], ds.lines)])
	total_lines_per_slice_at_ds.append(total_lines_per_slice_at_ds[-1])
	status_interval = {
		# twice per percent, but not too often or too seldom
		sliceno: min(max(total_lines_per_slice_at_ds[-1][sliceno] // 200, 10), 5000)
		for sliceno in want_slices
	}

	# never and always override env settings, auto (default) sets from env/tty
	if args.colour == 'never':
		colour.disable()
		highlight_matches = False
	elif args.colour == 'always':
		colour.enable()
		highlight_matches = True
	else:
		args.colour = 'auto'
		highlight_matches = colour.enabled

	# Don't highlight everything when just trying to cat
	if args.patterns == ['']:
		highlight_matches = False
	# Don't highlight anything with -l
	if args.list_matching:
		highlight_matches = False

	if args.format == 'json':
		# headers was just a mistake, ignore it
		args.headers = False

	separator = args.separator
	if args.tab_length:
		separator = None
	elif separator is None and not sys.stdout.isatty():
		separator = '\t'

	if separator is None:
		# special case where we try to be like a tab, but with spaces.
		# this is useful because terminals typically don't style tabs.
		# and also so you can use smarter tabs and change the length.
		def separate(items, lens, tab_len=tab_length.tab_len, field_len=tab_length.field_len, min_len=tab_length.min_len):
			things = []
			current_pos = min_pos = 0
			for item, item_len in zip(items, lens):
				things.append(item)
				current_pos += item_len
				min_pos += field_len
				spaces = max(min_pos - current_pos, min_len)
				align = (current_pos + spaces) % tab_len
				if align:
					spaces += tab_len - align
				things.append(colour(' ' * spaces, 'grep/separator'))
				current_pos += spaces
			return ''.join(things[:-1])
		separator = '\t'
	else:
		separator_coloured = colour(separator, 'grep/separator')
		def separate(items, lens):
			return separator_coloured.join(items)

	def json_default(obj):
		if isinstance(obj, (datetime.datetime, datetime.date, datetime.time)):
			return str(obj)
		elif isinstance(obj, complex):
			return [obj.real, obj.imag]
		else:
			return repr(obj)

	if args.format == 'csv':
		def escape_item(item):
			if item and (separator in item or item[0] in '\'"' or item[-1] in '\'"'):
				return '"' + item.replace('\n', '\\n').replace('"', '""') + '"'
			else:
				return item.replace('\n', '\\n')
		errors = 'surrogatepass'
	else:
		escape_item = None
		errors = 'replace' if PY2 else 'surrogateescape'

	# This is for the ^T handling. Each slice sends an update when finishing
	# a dataset, and every status_interval[sliceno] lines while iterating.
	# To minimise the data sent the only information sent over the queue
	# is (sliceno, finished_dataset).
	# Status printing is triggered by ^T (or SIGINFO if that is available)
	# or by SIGUSR1.
	# Pressing it again within two seconds prints stats per slice too.
	q_status = mp.LockFreeQueue()
	def status_collector():
		q_status.make_reader()
		status = {sliceno: [0, 0] for sliceno in want_slices}
		#            [ds_ix, done_lines]
		total_lines = sum(total_lines_per_slice_at_ds[-1])
		previous = [0]
		# base colour conf in if stderr is a tty, not stdout.
		if args.colour == 'auto':
			colour.configure_from_environ(stdout=sys.stderr)
		def show(sig, frame):
			t = monotonic()
			verbose = (previous[0] + 2 > t) # within 2 seconds of previous
			previous[0] = t
			ds_ixes = []
			progress_lines = []
			progress_fraction = []
			for sliceno in want_slices:
				ds_ix, done_lines = status[sliceno]
				ds_ixes.append(ds_ix)
				max_possible = min(done_lines + status_interval[sliceno], total_lines_per_slice_at_ds[ds_ix + 1][sliceno])
				done_lines = (done_lines + max_possible) / 2 # middle of the possibilities
				progress_lines.append(done_lines)
				total = total_lines_per_slice_at_ds[-1][sliceno]
				if total == 0:
					progress_fraction.append(1)
				else:
					progress_fraction.append(done_lines / total)
			progress_total = sum(progress_lines) / (total_lines or 1)
			bad_cutoff = progress_total - 0.1
			if verbose:
				show_ds = (len(datasets) > 1 and min(ds_ixes) != max(ds_ixes))
				for sliceno, ds_ix, p in zip(want_slices, ds_ixes, progress_fraction):
					if ds_ix == len(datasets):
						msg = 'DONE'
					else:
						msg = '{0:d}% of {1:n} lines'.format(round(p * 100), total_lines_per_slice_at_ds[-1][sliceno])
						if show_ds:
							msg = '%s (in %s)' % (msg, datasets[ds_ix].quoted,)
					msg = '%9d: %s' % (sliceno, msg,)
					if p < bad_cutoff:
						msg = colour(msg, 'grep/infohighlight')
					else:
						msg = colour(msg, 'grep/info')
					write(2, msg.encode('utf-8') + b'\n')
			msg = '{0:d}% of {1:n} lines'.format(round(progress_total * 100), total_lines)
			if len(datasets) > 1:
				min_ds = min(ds_ixes)
				max_ds = max(ds_ixes)
				if min_ds < len(datasets):
					ds_name = datasets[min_ds].quoted
					extra = '' if min_ds == max_ds else ' ++'
					msg = '%s (in %s%s)' % (msg, ds_name, extra,)
			worst = min(progress_fraction)
			if worst < bad_cutoff:
				msg = '%s, worst %d%%' % (msg, round(worst * 100),)
			msg = colour('  SUMMARY: %s' % (msg,), 'grep/info')
			write(2, msg.encode('utf-8') + b'\n')
		for signame in ('SIGINFO', 'SIGUSR1'):
			if hasattr(signal, signame):
				sig = getattr(signal, signame)
				signal.signal(sig, show)
				if hasattr(signal, 'pthread_sigmask'):
					signal.pthread_sigmask(signal.SIG_UNBLOCK, {sig})
		tc_original = None
		using_stdin = False
		if not hasattr(signal, 'SIGINFO') and sys.stdin.isatty():
			# ^T wont work automatically on this OS, so we need to handle it as terminal input
			import termios
			from accelerator.compat import selectors
			sel = selectors.DefaultSelector()
			sel.register(0, selectors.EVENT_READ)
			sel.register(q_status.r, selectors.EVENT_READ)
			try:
				tc_original = termios.tcgetattr(0)
				tc_changed = list(tc_original)
				tc_changed[3] &= ~(termios.ICANON | termios.IEXTEN)
				termios.tcsetattr(0, termios.TCSADRAIN, tc_changed)
				using_stdin = True
			except Exception:
				pass
			# we can't set stdin nonblocking, because it's probably the same
			# file description as stdout, so work around that with alarms.
			def got_alarm(sig, frame):
				raise IOError()
			signal.signal(signal.SIGALRM, got_alarm)
		try:
			while True:
				if using_stdin:
					do_q = False
					for key, _ in sel.select():
						if key.fd == 0:
							try:
								signal.alarm(1) # in case something else read it we block for max 1 second
								try:
									pressed = ord(os.read(0, 1))
								finally:
									signal.alarm(0)
								if pressed == 20:
									write(2, b'\n') # "^T" shows in the terminal
									os.kill(os.getpid(), signal.SIGUSR1)
							except Exception:
								pass
						elif key.fd == q_status.r:
							do_q = True
					if not do_q:
						continue
				try:
					sliceno, finished_dataset = q_status.get()
				except QueueEmpty:
					return
				if finished_dataset:
					ds_ix = status[sliceno][0] + 1
					status[sliceno] = [ds_ix, total_lines_per_slice_at_ds[ds_ix][sliceno]]
				else:
					status[sliceno][1] += status_interval[sliceno]
		finally:
			if tc_original is not None:
				try:
					termios.tcsetattr(0, termios.TCSADRAIN, tc_original)
				except Exception:
					pass
	status_process = mp.SimplifiedProcess(target=status_collector, name='ax grep status')
	# everything else will write, so make it a writer right away
	q_status.make_writer()

	# Output is only allowed while holding this lock, so that long lines
	# do not get intermixed. (Or when alone in producing output.)
	io_lock = Lock()

	# This contains some extra stuff to be a better base for the other
	# outputters.
	# When used directly it enforces no ordering, but merges smaller writes
	# to keep the number of syscalls down.

	class Outputter:
		def __init__(self, q_in, q_out):
			self.q_in = q_in
			self.q_out = q_out
			self.buffer = []
			self.merge_buffer = b''

		def put(self, data):
			self.merge_buffer += data
			if len(self.merge_buffer) >= 1024:
				self.move_merge()

		def move_merge(self):
			if self.merge_buffer:
				with io_lock:
					write(1, self.merge_buffer)
				self.merge_buffer = b''

		def start(self, ds):
			pass

		def end(self, ds):
			self.move_merge()

		def finish(self):
			pass

		def full(self):
			return len(self.buffer) > 5000

		def excite(self):
			self.move_merge()
			if self.buffer:
				self.pump(False)

	# Partially ordered output, each header change acts as a fence.
	# This is used in all slices except the first.
	#
	# The queue gets True when the previous slice is ready for the next
	# header change, and None when the header is printed (and it's ok
	# to resume output).

	class HeaderWaitOutputter(Outputter):
		def start(self, ds):
			expect_ds, show = next(show_headers_here)
			assert ds == expect_ds
			if show:
				self.add_wait()
			else:
				self.excite()

		def add_wait(self):
			# Each sync point is separated by None in the buffer
			self.buffer.append(None)
			self.buffer.append(b'') # Avoid need for special case in .drain
			self.pump()

		def move_merge(self):
			data = self.merge_buffer
			self.merge_buffer = b''
			if self.buffer:
				self.pump()
				if self.buffer:
					self.buffer.append(data)
					return
			with io_lock:
				write(1, data)

		def pump(self, wait=None):
			if wait is None:
				wait = self.full()
			try:
				got = self.q_in.get(wait)
			except QueueEmpty:
				if wait:
					# previous slice has exited without sending all messages
					raise
				return
			if got is True:
				# since pump is only called when we have outputted all
				# currently allowed output or when the next message is an
				# unblock for such output we can just unconditionally send
				# the True on to the next slice here.
				self.q_out.put(True)
				self.pump(wait)
				return
			else:
				self.q_out.put(None)
				self.drain()

		def drain(self):
			assert self.buffer[0] is None, 'The buffer must always stop at a sync point (or empty)'
			with io_lock:
				for pos, data in enumerate(self.buffer[1:], 1):
					if data is None:
						break
					elif data:
						write(1, data)
				else:
					# We did not reach the next fence, so last item is real data
					# and needs to be removed. (The buffer will then be empty and
					# output will continue directly until reaching the sync point.)
					pos += 1
			self.buffer[:pos] = ()

		def finish(self):
			while self.buffer:
				self.pump(True)

	# Partially ordered output, each header change acts as a fence.
	# This is used only in the first slice, and outputs the headers.
	#
	# When it is ready to output headers it sends True in the queue.
	# When the True has travelled around the queue ring all slices are
	# ready, the headers are printed, and None is sent to let the other
	# slices resume output.
	# (When the None returns it is ignored, because output is resumed
	# as soon as the headers are printed.)

	class HeaderOutputter(HeaderWaitOutputter):
		def add_wait(self):
			if not self.buffer:
				self.q_out.put(True)
			self.buffer.append(None)
			self.buffer.append(b'') # Avoid need for special case in .drain/.put
			self.pump()

		def drain(self):
			assert self.buffer[0] is None, 'The buffer must always stop at a sync point (or empty)'
			with io_lock:
				for pos, data in enumerate(self.buffer[1:], 1):
					if data is None:
						self.q_out.put(True)
						break
					elif data:
						write(1, data)
				else:
					pos += 1
			self.buffer[:pos] = ()

		def pump(self, wait=None):
			if wait is None:
				wait = self.full()
			try:
				got = self.q_in.get(wait)
			except QueueEmpty:
				if wait:
					# previous slice has exited without sending all messages
					raise
				return
			if got is True:
				# The True we put in when reaching the fence has travelled
				# all the way around the queue ring, it's time to print the
				# new headers
				write(1, next(headers_iter))
				# and then unblock the other slices
				self.q_out.put(None)
				self.drain()
				# No else, when the None comes back we just drop it.
			if not wait:
				self.pump(False)

	# Fully ordered output, each slice waits for the previous slice.
	# For each ds, waits for None (anything really) before starting,
	# sends None when done.

	class OrderedOutputter(Outputter):
		def start(self, ds):
			# Each ds is separated by None in the buffer
			self.buffer.append(None)
			self.buffer.append(b'') # Avoid need for special case in .drain
			self.pump()

		def end(self, ds):
			self.move_merge()
			if not self.buffer:
				# We are done with this ds, so let next slice continue
				self.q_out.put(None)

		def pump(self, wait=None):
			if wait is None:
				wait = self.full()
			try:
				self.q_in.get(wait)
			except QueueEmpty:
				if wait:
					# previous slice has exited without sending all messages
					raise
				return
			self.drain()

		def move_merge(self):
			data = self.merge_buffer
			self.merge_buffer = b''
			if self.buffer:
				self.pump()
				if self.buffer:
					self.buffer.append(data)
					return
			# No need for a lock, the other slices aren't writing concurrently.
			write(1, data)

		def drain(self):
			assert self.buffer[0] is None
			for pos, data in enumerate(self.buffer[1:], 1):
				if data is None:
					# We are done with this ds, so let next slice continue
					self.q_out.put(None)
					break
				elif data:
					write(1, data)
			else:
				# We did not reach the next ds, so last item is real data and
				# needs to be removed. (The buffer will then be empty and
				# output will continue directly until reaching the next ds.)
				pos += 1
			self.buffer[:pos] = ()

		def finish(self):
			not_finished = bool(self.buffer)
			while self.buffer:
				self.pump(True)
			if not_finished:
				self.q_out.put(None)

	# Same as above but for the first slice so it prints headers when needed.

	class OrderedHeaderOutputter(OrderedOutputter):
		def start(self, ds):
			# Each ds is separated by None in the buffer
			self.buffer.append(None)
			if headers:
				expect_ds, show = next(show_headers_here)
				assert ds == expect_ds
			else:
				show = False
			if show:
				# Headers changed, start with those.
				self.buffer.append(next(headers_iter))
			else:
				self.buffer.append(b'') # Avoid need for special case in .drain
			self.pump()

	# Choose the right outputter for the kind of sync we need.
	def outputter(q_in, q_out, first_slice=False):
		if args.list_matching:
			cls = Outputter
		elif args.ordered:
			if first_slice:
				cls = OrderedHeaderOutputter
			else:
				cls = OrderedOutputter
		elif headers:
			if first_slice:
				cls = HeaderOutputter
			else:
				cls = HeaderWaitOutputter
		else:
			cls = Outputter
		return cls(q_in, q_out)

	# Make printer for the selected output options
	def make_show(prefix, used_columns):
		def matching_ranges(item):
			ranges = []
			for p in patterns:
				ranges.extend(m.span() for m in p.finditer(item))
			if not ranges:
				return
			# merge overlapping/adjacent ranges
			ranges.sort()
			ranges = iter(ranges)
			start, stop = next(ranges)
			for a, b in ranges:
				if a <= stop:
					stop = max(stop, b)
				else:
					yield start, stop
					start, stop = a, b
			yield start, stop
		def filter_item(item):
			return ''.join(item[a:b] for a, b in matching_ranges(item))
		if args.format == 'json':
			dumps = json.JSONEncoder(ensure_ascii=False, default=json_default).encode
			def show(lineno, items):
				if only_matching == 'part':
					items = [filter_item(unicode(item)) for item in items]
				if only_matching == 'columns':
					d = {k: v for k, v in zip(used_columns, items) if filter_item(unicode(v))}
				else:
					d = dict(zip(used_columns, items))
				if args.show_lineno:
					prefix['lineno'] = lineno
				if prefix:
					prefix['data'] = d
					d = prefix
				return dumps(d).encode('utf-8', 'surrogatepass') + b'\n'
		else:
			def colour_item(item):
				pos = 0
				parts = []
				for a, b in matching_ranges(item):
					parts.extend((item[pos:a], colour(item[a:b], 'grep/highlight')))
					pos = b
				parts.append(item[pos:])
				return ''.join(parts)
			def show(lineno, items):
				data = list(prefix)
				if args.show_lineno:
					data.append(unicode(lineno))
				show_items = map(unicode, items)
				if only_matching:
					if only_matching == 'columns':
						show_items = (item if filter_item(item) else '' for item in show_items)
					else:
						show_items = map(filter_item, show_items)
				show_items = list(show_items)
				lens = (len(item) for item in data + show_items)
				if highlight_matches:
					show_items = list(map(colour_item, show_items))
				data.extend(show_items)
				if escape_item:
					lens_unesc = (len(item) for item in data)
					data = list(map(escape_item, data))
					lens_esc = (len(item) for item in data)
					lens = (l + esc - unesc for l, unesc, esc in zip(lens, lens_unesc, lens_esc))
				return separate(data, lens).encode('utf-8', errors) + b'\n'
		return show

	# This is called for each slice in each dataset.
	# Each slice has a separate process (the same for all datasets).
	# The first slice runs in the main process (unless -l), everything
	# else runs from one_slice.

	def grep(ds, sliceno, out):
		out.start(ds)
		if len(patterns) == 1:
			chk = patterns[0].search
		else:
			def chk(s):
				return any(p.search(s) for p in patterns)
		first = [True]
		def mk_iter(col):
			kw = {}
			if first[0]:
				first[0] = False
				lines = ds.lines[sliceno]
				if lines > status_interval[sliceno]:
					def cb(n):
						q_status.put((sliceno, False))
						out.excite()
					kw['callback'] = cb
					kw['callback_interval'] = status_interval[sliceno]
			if ds.columns[col].type == 'ascii':
				kw['_type'] = 'unicode'
			it = ds._column_iterator(sliceno, col, **kw)
			if ds.columns[col].type == 'bytes':
				errors = 'replace' if PY2 else 'surrogateescape'
				if ds.columns[col].none_support:
					it = (None if v is None else v.decode('utf-8', errors) for v in it)
				else:
					it = (v.decode('utf-8', errors) for v in it)
			return it
		used_columns = columns_for_ds(ds)
		used_grep_columns = grep_columns and columns_for_ds(ds, grep_columns)
		if grep_columns and set(used_grep_columns) != set(used_columns):
			grep_iter = izip(*(mk_iter(col) for col in used_grep_columns))
		else:
			grep_iter = repeat(None)
		lines_iter = izip(*(mk_iter(col) for col in used_columns))
		if args.before_context:
			before = deque((), args.before_context)
		else:
			before = None
		if args.format == 'json':
			prefix = {}
			if args.show_dataset:
				prefix['dataset'] = ds
			if args.show_sliceno:
				prefix['sliceno'] = sliceno
			show = make_show(prefix, used_columns)
		else:
			prefix = []
			if args.show_dataset:
				prefix.append(ds)
			if args.show_sliceno:
				prefix.append(str(sliceno))
			prefix = tuple(prefix)
			show = make_show(prefix, used_columns)
		if args.invert_match:
			maybe_invert = operator.not_
		else:
			maybe_invert = bool
		to_show = 0
		for lineno, (grep_items, items) in enumerate(izip(grep_iter, lines_iter)):
			if maybe_invert(any(chk(unicode(item)) for item in grep_items or items)):
				if q_list:
					q_list.put((ds, sliceno))
					return
				while before:
					out.put(show(*before.popleft()))
				to_show = 1 + args.after_context
			if to_show:
				out.put(show(lineno, items))
				to_show -= 1
			elif before is not None:
				before.append((lineno, items))
		out.end(ds)

	# This runs in a separate process for each slice except the first
	# one (unless -l), which is handled specially in the main process.

	def one_slice(sliceno, q_in, q_out, q_to_close):
		if q_to_close:
			q_to_close.close()
		if q_in:
			q_in.make_reader()
		if q_out:
			q_out.make_writer()
		if q_list:
			q_list.make_writer()
		try:
			out = outputter(q_in, q_out)
			for ds in datasets:
				if seen_list is None or ds not in seen_list:
					grep(ds, sliceno, out)
				q_status.put((sliceno, True))
			out.finish()
		except QueueEmpty:
			# some other process died, no need to print an error here
			sys.exit(1)

	headers_prefix = []
	if args.show_dataset:
		headers_prefix.append('[DATASET]')
	if args.show_sliceno:
		headers_prefix.append('[SLICE]')
	if args.show_lineno:
		headers_prefix.append('[LINE]')

	# [headers] for each ds where headers change (not including the first).
	# this is every ds where sync between slices has to happen when not --ordered.
	# which ds this is is stored in show_headers_here
	headers = []
	if args.headers:
		current_headers = None
		# [(ds, show_headers?)] for each ds in datasets.
		# records ds as well to uncover bugs (it's not really needed)
		show_headers_here = []
		for ds in datasets:
			candidate_headers = columns_for_ds(ds)
			if candidate_headers != current_headers:
				current_headers = candidate_headers
				headers.append(current_headers)
				show_headers_here.append((ds, True))
			else:
				show_headers_here.append((ds, False))
		def gen_headers(headers):
			show_items = headers_prefix + headers
			if escape_item:
				show_items = list(map(escape_item, show_items))
			coloured = (colour(item, 'grep/header') for item in show_items)
			txt = separate(coloured, map(len, show_items))
			return txt.encode('utf-8', 'surrogatepass') + b'\n'
		# remove the starting ds, so no header changes means no special handling.
		current_headers = headers.pop(0)
		if not args.list_matching:
			write(1, gen_headers(current_headers))
		headers_iter = iter(map(gen_headers, headers))
		show_headers_here[0] = (datasets[0], False) # first one is already printed
		show_headers_here = iter(show_headers_here)

	q_in = q_out = first_q_out = q_to_close = q_list = None
	children = [status_process]
	seen_list = None
	if args.list_matching:
		# in this case all slices get their own process
		# and the main process just prints the maching slices
		q_list = mp.LockFreeQueue()
		separate_process_slices = want_slices
		if not args.show_sliceno:
			seen_list = mp.MpSet()
	else:
		separate_process_slices = want_slices[1:]
		if args.ordered or headers:
			# needs to sync in some way
			q_in = first_q_out = mp.LockFreeQueue()
	for sliceno in separate_process_slices:
		if q_in:
			q_out = mp.LockFreeQueue()
		p = mp.SimplifiedProcess(
			target=one_slice,
			args=(sliceno, q_in, q_out, q_to_close,),
			name='slice-%d' % (sliceno,),
		)
		children.append(p)
		if q_in and q_in is not first_q_out:
			q_in.close()
		q_to_close = first_q_out
		q_in = q_out
	if q_in:
		if q_in is first_q_out:
			# Special case: only one slice, but HeaderOutputter still needs queues.
			class LocalQueue:
				def __init__(self):
					self._lst = []
					self.put = self._lst.append
				def get(self, wait=None):
					if self._lst:
						return self._lst.pop(0)
					raise QueueEmpty()
			q_in = q_out = LocalQueue()
		else:
			q_out = first_q_out
			q_in.make_reader()
			q_out.make_writer()
		if args.ordered:
			q_in.put_local(None)
	del q_to_close
	del first_q_out

	try:
		if args.list_matching:
			if args.headers:
				headers_prefix = ['[DATASET]']
				if seen_list is None:
					headers_prefix.append('[SLICE]')
				write(1, gen_headers([]))
			ordered_res = defaultdict(set)
			q_list.make_reader()
			if seen_list is None:
				used_columns = ['dataset', 'sliceno']
			else:
				used_columns = ['dataset']
			inner_show = make_show({} if args.format == 'json' else [], used_columns)
			def show(ds, sliceno=None):
				if sliceno is None:
					items = [ds]
				else:
					items = [ds, sliceno]
				write(1, inner_show(None, items))
			while True:
				try:
					ds, sliceno = q_list.get()
				except QueueEmpty:
					break
				if seen_list is None:
					if args.ordered:
						ordered_res[ds].add(sliceno)
					else:
						show(ds, sliceno)
				elif ds not in seen_list:
					seen_list.add(ds)
					if not args.ordered:
						show(ds)
			if args.ordered:
				for ds in datasets:
					if seen_list is None:
						for sliceno in sorted(ordered_res[ds]):
							show(ds, sliceno)
					else:
						if ds in seen_list:
							show(ds)
		else:
			out = outputter(q_in, q_out, first_slice=True)
			sliceno = want_slices[0]
			for ds in datasets:
				grep(ds, sliceno, out)
				q_status.put((sliceno, True))
			out.finish()
	except QueueEmpty:
		# don't print an error, probably a subprocess died from EPIPE before
		# the main process. (or the subprocess already printed an error.)
		return 1

	q_status.close()
	for c in children:
		c.join()
		if c.exitcode:
			return 1



############################################################################
#                                                                          #
# Copyright (c) 2021-2022 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division

import os
import sys
import select
import signal
import fcntl
import multiprocessing
import struct
import pickle
import errno
from traceback import print_exc
from accelerator.compat import QueueEmpty, monotonic, selectors


assert select.PIPE_BUF >= 512, "POSIX says PIPE_BUF is at least 512, you have %d" % (select.PIPE_BUF,)

PIPE_BUF = min(select.PIPE_BUF, 65536)
MAX_PART = PIPE_BUF - 6


def _nb(fd):
	fl = fcntl.fcntl(fd, fcntl.F_GETFL)
	fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)


def _mk_sel(fd, ev):
	sel = selectors.DefaultSelector()
	sel.register(fd, ev)
	return sel


# This provides some of the functionality of multiprocessing.Queue but
# without using any locks, instead relying on writes up to PIPE_BUF being
# atomic. No threads are used.
#
# Only one process can .get() (after .make_reader()). .get() can take a timeout.
# (There are no checks to enforce a single reader, so be careful.)
# Any number of processeses can .put(). .put() blocks until the pipe will take
# all the data.
# A dying writer will never corrupt or deadlock the queue, but can cause a
# memory leak in the reading process (until the queue object is destroyed).
# Large messages are probably quite inefficient.
# You can call .make_*() before forking, but once you have put anything in the
# queue it is no longer fork safe. (Forking is still fine, but not using the
# queue from the new process.)
#
# In short, you often can't use this. And when you do, you have to be careful.

class QueueClosedError(Exception):
	__slots__ = ()

class LockFreeQueue:
	__slots__ = ('r', 'w', '_r_sel', '_w_sel', '_pid', '_buf', '_partial', '_local')

	def __init__(self):
		self._r_sel = self._w_sel = None
		self.r, self.w = os.pipe()
		self._buf = b''
		self._partial = {}
		self._local = []

	def make_writer(self):
		os.close(self.r)
		self.r = -1

	def make_reader(self):
		os.close(self.w)
		self.w = -1

	def close(self):
		if self._r_sel:
			self._r_sel.close()
			self._r_sel = None
		if self._w_sel:
			self._w_sel.close()
			self._w_sel = None
		if self.r != -1:
			os.close(self.r)
			self.r = -1
		if self.w != -1:
			os.close(self.w)
			self.w = -1

	__del__ = close

	def _late_setup(self):
		# The selector must be constructed late so it happens in the right
		# process. Also probably saves an fd before fork (in the selector).
		pid = os.getpid()
		if self._r_sel or self._w_sel:
			assert self._pid == pid, "Can't fork after using"
		else:
			self._pid = pid
			if self.r != -1:
				_nb(self.r)
				self._r_sel = _mk_sel(self.r, selectors.EVENT_READ)
			if self.w != -1:
				_nb(self.w)
				self._w_sel = _mk_sel(self.w, selectors.EVENT_WRITE)

	def get(self, block=True, timeout=0):
		assert self.w == -1, "call make_reader first"
		if self._local:
			return self._local.pop(0)
		self._late_setup()
		if timeout:
			deadline = monotonic() + timeout
		need_data = False
		eof = (self.r == -1)
		while True:
			if not eof:
				try:
					data = os.read(self.r, PIPE_BUF)
					if not data:
						eof = True
					self._buf += data
					need_data = False
				except OSError:
					pass
			if len(self._buf) >= 6:
				z, pid = struct.unpack('<HI', self._buf[:6])
				assert pid, "all is lost"
				if len(self._buf) < 6 + z:
					need_data = True
				else:
					data = self._buf[6:6 + z]
					self._buf = self._buf[6 + z:]
					if pid not in self._partial:
						want_len = struct.unpack("<I", data[:4])[0]
						have = [data[4:]]
						have_len = len(have[0])
					else:
						want_len, have, have_len = self._partial[pid]
						have.append(data)
						have_len += len(data)
					if have_len == want_len:
						self._partial.pop(pid, None)
						data = b''.join(have)
						return pickle.loads(data)
					else:
						self._partial[pid] = (want_len, have, have_len)
			if len(self._buf) < 6 or need_data:
				if eof:
					raise QueueEmpty()
				if not block:
					if not self._r_sel.select(0):
						raise QueueEmpty()
				elif timeout:
					time_left = deadline - monotonic()
					if time_left <= 0:
						raise QueueEmpty()
					self._r_sel.select(time_left)
				else:
					self._r_sel.select()

	def put_local(self, msg):
		self._local.append(msg)

	def put(self, msg):
		assert self.r == -1, "call make_writer first"
		if self.w == -1:
			raise QueueClosedError()
		self._late_setup()
		msg = pickle.dumps(msg, pickle.HIGHEST_PROTOCOL)
		msg = struct.pack('<I', len(msg)) + msg
		offset = 0
		while offset < len(msg):
			part = msg[offset:offset + MAX_PART]
			part = struct.pack('<HI', len(part), self._pid) + part
			offset += MAX_PART
			while True:
				self._w_sel.select()
				try:
					wlen = os.write(self.w, part)
				except OSError as e:
					if e.errno == errno.EAGAIN:
						wlen = 0
					else:
						self.close()
						raise QueueClosedError()
				if wlen:
					if wlen != len(part):
						print("OS violates PIPE_BUF guarantees, all is lost.", file=sys.stderr)
						while True:
							try:
								# this should eventually make the other side read PID 0
								os.write(self.w, b'\0')
							except OSError:
								pass
					break

	def try_notify(self):
		pid = os.getpid()
		msg = struct.pack('<HII', 6, pid, 2) + b'N.' # pickled None
		try:
			os.write(self.w, msg)
			return True
		except OSError:
			return False


# This is a partial replacement for multiprocessing.Process.
# It doesn't work if you use the rest of the mp machinery (like Queues)
# and always uses os.fork(). It exists because multiprocessing.Process
# has scaling issues with many children.

class SimplifiedProcess:
	__slots__ = ('pid', 'name', '_alive', 'exitcode')

	def __init__(self, target, args=(), kwargs={}, name=None):
		sys.stdout.flush()
		sys.stderr.flush()
		self.pid = os.fork()
		if self.pid:
			self.name = name
			self._alive = True
			return
		rc = 1
		try:
			target(*args, **kwargs)
			rc = 0
		except KeyboardInterrupt:
			signal.signal(signal.SIGINT, signal.SIG_DFL)
			os.kill(os.getpid(), signal.SIGINT)
		except Exception as e:
			if not isinstance(e, OSError) or e.errno != errno.EPIPE:
				print("Exception in %d %r:" % (os.getpid(), name), file=sys.stderr)
				print_exc(file=sys.stderr)
		except SystemExit as e:
			if e.code is None:
				rc = 0
			elif isinstance(e.code, int):
				rc = e.code
			else:
				print(e.code, file=sys.stderr)
		finally:
			os._exit(rc)

	def is_alive(self):
		self._wait(False)
		return self._alive

	def _wait(self, block):
		if not self._alive:
			return
		pid, status = os.waitpid(self.pid, 0 if block else os.WNOHANG)
		if pid:
			assert pid == self.pid
			self._alive = False
			if os.WIFEXITED(status):
				self.exitcode = os.WEXITSTATUS(status)
			elif os.WIFSIGNALED(status):
				self.exitcode = -os.WTERMSIG(status)
			else:
				self.exitcode = -999

	def join(self):
		self._wait(True)


class MpSet:
	__slots__ = ('_lock', '_q_r', '_q_w', '_p', '_broken')

	def __init__(self):
		self._lock = multiprocessing.Lock()
		self._q_r = LockFreeQueue()
		self._q_w = LockFreeQueue()
		self._p = SimplifiedProcess(target=self._process, name='MpSet')
		self._q_r.make_reader()
		self._q_w.make_writer()
		self._broken = False

	def _process(self):
		fd_keep = {self._q_r.r, self._q_r.w, self._q_w.r, self._q_w.w}
		fd = 3
		fd_end = fd + 20
		while fd < fd_end:
			if fd not in fd_keep:
				try:
					os.close(fd)
					fd_end = fd + 20
				except OSError:
					pass
			fd += 1
		self._q_r.make_writer()
		self._q_w.make_reader()
		s = set()
		while True:
			try:
				funcname, a = self._q_w.get()
			except QueueEmpty:
				return
			try:
				res = getattr(s, funcname)(*a)
				err = None
			except Exception as e:
				res = None
				err = e
			self._q_r.put((res, err))

	def _call(self, funcname, *a):
		assert not self._broken
		self._lock.acquire()
		try:
			self._q_w.put((funcname, a))
			res, err = self._q_r.get()
		except:
			self._broken = True
			raise
		finally:
			self._lock.release()
		if err:
			raise err
		return res

	def add(self, value):
		return self._call('add', value)

	def __contains__(self, value):
		return self._call('__contains__', value)

	def __len__(self):
		return self._call('__len__')



############################################################################
#                                                                          #
# Copyright (c) 2020-2021 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

try:
	# setup.py needs to import some things, let's not break that.
	# When adding types to _dsutil switch this to one of those
	# (to make building with a previous version installed work).
	from ._dsutil import ReadComplex64
	del ReadComplex64
	before_install = False
except ImportError:
	before_install = True

__all__ = []

if not before_install:
	def get_version():
		import os.path
		try:
			dn = os.path.dirname(__file__)
			fn = os.path.join(dn, 'version.txt')
			with open(fn, 'r') as fh:
				return next(fh).strip()
		except Exception:
			return None
	__version__ = get_version()
	del get_version
	from .dataset import SkipSlice, SkipDataset, DatasetList
	__all__.extend((SkipSlice, SkipDataset, DatasetList,))
	from .error import AcceleratorError, UserError, ServerError
	from .error import UrdError, UrdPermissionError, UrdConflictError
	from .error import NoSuchWhateverError, NoSuchJobError, NoSuchWorkdirError
	from .error import DatasetError, NoSuchDatasetError, DatasetUsageError
	from .error import JobError, ColourError, BuildError
	__all__.extend((AcceleratorError, UserError, ServerError,))
	__all__.extend((UrdError, UrdPermissionError, UrdConflictError,))
	__all__.extend((NoSuchWhateverError, NoSuchJobError, NoSuchWorkdirError,))
	__all__.extend((DatasetError, NoSuchDatasetError, DatasetUsageError,))
	__all__.extend((JobError, ColourError, BuildError,))
	from .extras import DotDict
	from .extras import OptionEnum, OptionString, RequiredOption, OptionDefault
	__all__.extend((DotDict,))
	__all__.extend((OptionEnum, OptionString, RequiredOption, OptionDefault,))
	from .job import Job, JobWithFile
	__all__.extend((Job, JobWithFile,))
	from .statmsg import status, dummy_status
	__all__.extend((status, dummy_status,))
	from .subjobs import build
	__all__.extend((build,))
	from .colourwrapper import colour
	__all__.extend((colour,))
	# hack to make "from accelerator.colour import bold" and similar work.
	import sys
	sys.modules['accelerator.colour'] = colour
	del sys
	colour.Colour = type(colour) # so you can import the class too.
	__all__ = [k for k, v in locals().items() if v in __all__]

del before_install



############################################################################
#                                                                          #
# Copyright (c) 2021 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Create some jobs and datasets to test the shell parser with.

Creates a default dataset with previous from datasets.previous,
and a j dataset with previous from j in jobs.previous.
'''

datasets = ('previous', 'parent',)
jobs = ('previous',)

def synthesis(job):
	dw = job.datasetwriter(columns={'a': 'bool'}, previous=datasets.previous)
	dw.get_split_write()
	previous = jobs.previous.dataset('j') if jobs.previous else None
	dw = job.datasetwriter(name='j', columns={'b': 'bool'}, previous=previous, parent=datasets.parent)
	dw.get_split_write()
	dw = job.datasetwriter(name='name/with/slash', columns={'b': 'bool'})
	dw.get_split_write()



############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2018-2021 Carl Drougge                       #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import division
from __future__ import absolute_import

description = r'''Dataset (or chain) to CSV file.'''

from shutil import copyfileobj
from os import unlink
from os.path import exists
from contextlib import contextmanager
from json import JSONEncoder
from functools import partial
from itertools import chain
import gzip

from accelerator.compat import PY3, PY2, izip, imap, long
from accelerator import status


options = dict(
	filename          = 'result.csv', # .gz or with .gz. in it for compressed mode
	separator         = ',',
	line_separator    = '\n',
	labelsonfirstline = True,
	chain_source      = False, # everything in source is replaced by datasetchain(self, stop=from previous)
	quote_fields      = '', # can be any string, but use '"' or "'"
	lazy_quotes       = True, # only quote field if value needs quoting
	none_as           = None, # A string or {label: string} to use for None-values. Default 'None' ('null' for json).
	labels            = [], # empty means all labels in (first) dataset
	sliced            = False, # one output file per slice, you can put %02d or similar in filename (or get filename.%d)
	compression       = 6,     # gzip level
)

datasets = (['source'],) # normally just one, but you can specify several

jobs = ('previous',)

@contextmanager
def writer(fh):
	with fh:
		write = fh.write
		line_sep = options.line_separator
		def wrapped_write(s):
			write(s)
			write(line_sep)
		yield wrapped_write

format = dict(
	ascii=None,
	float32=repr,
	float64=repr,
	json=JSONEncoder(sort_keys=True, ensure_ascii=True, check_circular=False).encode,
)

if PY3:
	enc = str
	format['bytes'] = lambda s: s.decode('utf-8', errors='backslashreplace')
	format['number'] = repr
	format['unicode'] = None
else:
	enc = lambda s: s.encode('utf-8')
	format['bytes'] = None
	format['number'] = lambda n: str(n) if isinstance(n, long) else repr(n)
	format['unicode'] = lambda s: s.encode('utf-8')

def csvexport(sliceno, filename, labelsonfirstline):
	d = datasets.source[0]
	if not options.labels:
		options.labels = sorted(d.columns)
	if options.chain_source:
		if jobs.previous:
			prev_source = jobs.previous.params.datasets.source
			assert len(datasets.source) == len(prev_source)
		else:
			prev_source = [None] * len(datasets.source)
		lst = []
		for src, stop in zip(datasets.source, prev_source):
			lst.extend(src.chain(stop_ds=stop))
		datasets.source = lst
	if options.filename.lower().endswith('.gz') or '.gz.' in options.filename.lower():
		open_func = partial(gzip.open, compresslevel=options.compression)
	else:
		open_func = open
	if PY2:
		open_func = partial(open_func, mode='wb')
	else:
		open_func = partial(open_func, mode='xt', encoding='utf-8')
	if options.none_as:
		if isinstance(options.none_as, dict):
			bad_none = set(options.none_as) - set(options.labels)
			assert not bad_none, 'Unknown labels in none_as: %r' % (bad_none,)
		else:
			assert isinstance(options.none_as, str), "What did you pass as none_as?"
	def resolve_none(label, col):
		d = options.none_as or {}
		if col.type in ('json', 'pickle',):
			if isinstance(options.none_as, str):
				return options.none_as
			return d.get(label)
		elif col.none_support:
			if isinstance(options.none_as, str):
				return options.none_as
			return d.get(label, 'None')
	q = options.quote_fields
	qq = q + q
	sep = options.separator
	def quote_always(v):
		return q + v.replace(q, qq) + q
	if q in '"\'':
		# special case so both quotes will quote the other
		def quote_if_needed(v):
			if v and (v[0] in '"\'' or v[-1] in '"\'' or sep in v):
				return q + v.replace(q, qq) + q
			else:
				return v
	else:
		def quote_if_needed(v):
			if v.startswith(q) or v.endswith(q) or sep in v:
				return q + v.replace(q, qq) + q
			else:
				return v
	if not q:
		quote_func = str
	elif options.lazy_quotes and sep: # always quote if no separator
		quote_func = quote_if_needed
	else:
		quote_func = quote_always
	def needs_quoting(typ):
		if not q:
			return False
		if not options.lazy_quotes:
			return True
		# maybe we can skip quoting because values that need quoting are impossible?
		if typ in ('int32', 'int64', 'bits32', 'bits64',):
			possible = '0123456789-'
		elif typ in ('float32', 'float64', 'number',):
			possible = '0123456789-+einfa.'
		else:
			possible = False
		if possible:
			q_s = set(q)
			sep_s = set(sep)
			possible_s = set(possible)
			if q_s - possible_s and sep_s - possible_s:
				return False
		return True
	def column_iterator(d, label, first):
		col = d.columns[label]
		f = format.get(col.type, str)
		it = d.iterate(sliceno, label, status_reporting=first)
		none_as = resolve_none(label, col)
		if none_as is not None:
			none_as = quote_func(none_as)
			if needs_quoting(col.type):
				if f:
					it = (none_as if v is None else quote_func(f(v)) for v in it)
				else:
					it = (none_as if v is None else quote_func(v) for v in it)
			else:
				if f:
					it = (none_as if v is None else f(v) for v in it)
				else:
					it = (none_as if v is None else v for v in it)
		elif f:
			if needs_quoting(col.type):
				it = (quote_func(f(v)) for v in it)
			else:
				it = imap(f, it)
		elif needs_quoting(col.type):
			it = imap(quote_func, it)
		return it
	def outer_iterator(label, first):
		return chain.from_iterable(column_iterator(d, label, first) for d in datasets.source)
	iters = []
	first = True
	for label in options.labels:
		iters.append(outer_iterator(label, first))
		first = False
	it = izip(*iters)
	with writer(open_func(filename)) as write:
		if labelsonfirstline:
			write(enc(sep.join(map(quote_func, options.labels))))
		for data in it:
			write(sep.join(data))

def analysis(sliceno, job):
	if options.sliced:
		if '%' in options.filename:
			filename = options.filename % (sliceno,)
		else:
			filename = '%s.%d' % (options.filename, sliceno,)
		csvexport(sliceno, filename, options.labelsonfirstline)
		job.register_file(filename)
	else:
		labelsonfirstline = (sliceno == 0 and options.labelsonfirstline)
		csvexport(sliceno, str(sliceno), labelsonfirstline)

def synthesis(job, slices):
	if not options.sliced:
		def msg(sliceno):
			return "Assembling %s (%d/%d)" % (options.filename, sliceno + 1, slices,)
		with status(msg(0)) as update:
			with job.open(options.filename, "wb") as outfh:
				for sliceno in range(slices):
					filename = str(sliceno)
					if exists(filename):
						update(msg(sliceno))
						with open(filename, "rb") as infh:
							copyfileobj(infh, outfh)
						unlink(filename)
